from __future__ import annotations

import itertools
import random
import time
import warnings
from collections import defaultdict
from pathlib import Path
from time import sleep
from typing import Sequence

import hypothesis.strategies as st
import numpy as np
import pkg_resources
import pytest
import pytest_check as check
from hypothesis import HealthCheck, example, given, settings
from pytest import mark

import negmas
from negmas import SAOSyncController
from negmas.gb.negotiators.randneg import RandomAlwaysAcceptingNegotiator
from negmas.genius import genius_bridge_is_running
from negmas.helpers import unique_name
from negmas.helpers.strings import shorten
from negmas.helpers.types import get_class
from negmas.outcomes import Outcome, make_issue
from negmas.outcomes.outcome_space import make_os
from negmas.preferences import LinearUtilityFunction, MappingUtilityFunction
from negmas.sao import (
    AdditiveParetoFollowingTBNegotiator,
    AspirationNegotiator,
    ConcederTBNegotiator,
    LimitedOutcomesNegotiator,
    MultiplicativeParetoFollowingTBNegotiator,
    RandomNegotiator,
    ResponseType,
    SAOMechanism,
    SAONegotiator,
    SAOResponse,
    SAOState,
    all_negotiator_types,
)
from negmas.sao.negotiators import NaiveTitForTatNegotiator, NiceNegotiator

exception_str = "Custom Exception"

NEGTYPES = all_negotiator_types()

TIME_BASED_NEGOTIATORS = [
    get_class(f"negmas.sao.negotiators.timebased.{x}")
    for x in negmas.sao.negotiators.timebased.__all__
]
TFT_NEGOTIATORS = [
    get_class(f"negmas.sao.negotiators.titfortat.{x}")
    for x in negmas.sao.negotiators.titfortat.__all__
]
ALL_BUILTIN_NEGOTIATORS = [
    get_class(f"negmas.sao.negotiators.{x}")
    for x in negmas.sao.negotiators.__all__
    if x
    not in [
        "SAONegotiator",
        "UtilBasedNegotiator",
        "make_boa",
        "MakeBoa",
        "SAOModularNegotiator",
        "BOANegotiator",
        "ControlledSAONegotiator",
        "MAPNegotiator",
        "CABNegotiator",
        "CANNegotiator",
        "CARNegotiator",
        "WABNegotiator",
        "WANNegotiator",
        "WARNegotiator",
    ]
]


class MyRaisingNegotiator(SAONegotiator):
    def propose(self, state: SAOState) -> Outcome | None:
        _ = state
        raise ValueError(exception_str)


class MySyncController(SAOSyncController):
    def __init__(
        self,
        *args,
        sleep_seconds=0.2,
        accept_after=float("inf"),
        end_after=float("inf"),
        offer_none_after=float("inf"),
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._sleep_seconds = sleep_seconds
        self.n_counter_all_calls = 0
        self.countered_offers: dict[int, dict[str, list[Outcome | None]]] = defaultdict(
            lambda: defaultdict(list)
        )
        self.received_offers: dict[str, dict[int, list[Outcome | None]]] = defaultdict(
            lambda: defaultdict(list)
        )
        self.sent_offers: dict[str, dict[int, list[Outcome | None]]] = defaultdict(
            lambda: defaultdict(list)
        )
        self.sent_responses: dict[str, dict[int, list[ResponseType]]] = defaultdict(
            lambda: defaultdict(list)
        )
        self.wait_states: dict[str, dict[int, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        self.accept_after = accept_after
        self.end_after = end_after
        self.offer_none_after = offer_none_after

    def respond(self, negotiator_id, state, source: str | None = None):
        offer = state.current_offer
        response = super().respond(negotiator_id, state, source)
        self.received_offers[negotiator_id][state.step].append(offer)
        if response == ResponseType.WAIT:
            self.wait_states[negotiator_id][state.step] += 1
        else:
            self.sent_responses[negotiator_id][state.step].append(response)
        return response

    def propose(self, negotiator_id, state):
        outcome = super().propose(negotiator_id, state)
        self.sent_offers[negotiator_id][state.step].append(outcome)
        return outcome

    def counter_all(self, offers, states):
        for k, v in offers.items():
            s = states[k]
            self.countered_offers[s.step][k].append(v)

        if self._sleep_seconds:
            if isinstance(self._sleep_seconds, Sequence):
                s = (
                    random.random() * (self._sleep_seconds[1] - self._sleep_seconds[0])
                    + self._sleep_seconds[0]
                )
            else:
                s = self._sleep_seconds
            sleep(s)
        self.n_counter_all_calls += 1
        responses = [
            self.negotiators[_][0].nmi.random_outcomes(1)[0] for _ in states.keys()
        ]
        assert all(_ is not None for _ in responses)
        responses = dict(
            zip(
                states.keys(),
                (SAOResponse(ResponseType.REJECT_OFFER, _) for _ in responses),
            )
        )
        if (
            self.offer_none_after <= self.end_after
            and self.offer_none_after <= self.accept_after
        ):
            for k in responses.keys():
                if states[k].step >= self.offer_none_after:
                    responses[k] = SAOResponse(ResponseType.REJECT_OFFER, None)
        for k in responses.keys():
            if self.accept_after < self.end_after:
                if states[k].step >= self.accept_after:
                    responses[k] = SAOResponse(ResponseType.ACCEPT_OFFER, None)
                if states[k].step >= self.end_after:
                    responses[k] = SAOResponse(ResponseType.END_NEGOTIATION, None)
            else:
                if states[k].step >= self.end_after:
                    responses[k] = SAOResponse(ResponseType.END_NEGOTIATION, None)
                if states[k].step >= self.accept_after:
                    responses[k] = SAOResponse(ResponseType.ACCEPT_OFFER, None)
        return responses

    def first_offer(self, negotiator_id: str):
        return self.negotiators[negotiator_id][0].nmi.random_outcomes(1)[0]


class InfiniteLoopNegotiator(RandomNegotiator):
    """A negotiator that wastes time"""

    def __init__(self, *args, **kwargs):
        pa, pe, pr = 0.0, 0.0, 1.0
        kwargs["p_acceptance"] = kwargs.get("p_acceptance", pa)
        kwargs["p_ending"] = kwargs.get("p_ending", pe)
        kwargs["p_rejection"] = kwargs.get("p_rejection", pr)
        super().__init__(*args, **kwargs)
        self.__stop = False

    def __call__(self, state):
        _ = state
        while not self.__stop:
            pass

    def stop(self):
        self.__stop = True


class TimeWaster(RandomNegotiator):
    """A negotiator that wastes time"""

    def __init__(
        self,
        *args,
        sleep_seconds: float | tuple[float, float] = 0.2,
        n_waits=0,
        **kwargs,
    ):
        pa, pe, pr = 0.0, 0.0, 1.0
        kwargs["p_acceptance"] = kwargs.get("p_acceptance", pa)
        kwargs["p_ending"] = kwargs.get("p_ending", pe)
        kwargs["p_rejection"] = kwargs.get("p_rejection", pr)
        super().__init__(*args, **kwargs)
        self._sleep_seconds = sleep_seconds
        self.my_offers: dict[int, Outcome | None] = defaultdict(lambda: None)
        self.received_offers: dict[int, Outcome | None] = defaultdict(lambda: None)
        self.my_responses: dict[int, ResponseType | None] = defaultdict(lambda: None)
        self.n_waits = n_waits
        if n_waits:
            self.n_waits = random.randint(0, n_waits)
        self.waited = 0

    def __call__(self, state):
        offer = state.current_offer
        if not self.nmi:
            return None
        if self.waited < self.n_waits and state.step > 0:
            self.waited += 1
            return SAOResponse(ResponseType.WAIT, None)
        if self._sleep_seconds:
            if isinstance(self._sleep_seconds, Sequence):
                s = (
                    random.random() * (self._sleep_seconds[1] - self._sleep_seconds[0])
                    + self._sleep_seconds[0]
                )
            else:
                s = self._sleep_seconds
            sleep(s)
        assert (
            state.step not in self.received_offers
            or self.received_offers[state.step] is None
        )
        assert state.step not in self.my_offers
        assert state.step not in self.my_responses
        self.received_offers[state.step] = offer
        outcome = self.nmi.random_outcome()
        response = SAOResponse(ResponseType.REJECT_OFFER, outcome)
        assert response.outcome is not None
        self.my_offers[state.step] = response.outcome
        self.my_responses[state.step] = response.response
        return response


def test_hidden_time_limit_words():
    n_outcomes, n_steps, tlimit = 100, 10, 1
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=n_steps,
        ignore_negotiator_exceptions=True,
        hidden_time_limit=tlimit,
        step_time_limit=float("inf"),
    )
    ufuns = MappingUtilityFunction.generate_random(
        2, outcomes=mechanism.discrete_outcomes()
    )
    mechanism.add(InfiniteLoopNegotiator(name=f"agent{0}", preferences=ufuns[0]))  # type: ignore
    mechanism.add(InfiniteLoopNegotiator(name=f"agent{1}", preferences=ufuns[1]))  # type: ignore
    mechanism.run()
    assert mechanism.state.agreement is None
    assert mechanism.state.started
    assert mechanism.state.timedout
    assert mechanism.state.step < n_steps
    assert mechanism._start_time is not None
    assert time.perf_counter() - mechanism._start_time >= tlimit
    assert not mechanism.state.waiting
    for negotiator in mechanism.negotiators:
        negotiator.stop()


def test_neg_run_no_waiting():
    n_outcomes, n_steps, waste = 10, 10, 0.5
    mechanism = SAOMechanism(
        outcomes=n_outcomes, n_steps=n_steps, ignore_negotiator_exceptions=True
    )
    ufuns = MappingUtilityFunction.generate_random(
        2, outcomes=mechanism.discrete_outcomes()
    )
    mechanism.add(
        TimeWaster(name=f"agent{0}", sleep_seconds=waste, preferences=ufuns[0])
    )
    mechanism.add(
        TimeWaster(name=f"agent{1}", sleep_seconds=waste, preferences=ufuns[1])
    )
    mechanism.run()
    assert mechanism.state.agreement is None
    assert mechanism.state.started
    assert mechanism.state.timedout
    assert mechanism.state.step == n_steps
    assert not mechanism.state.waiting
    assert len(mechanism.history) == n_steps
    for _, v in mechanism.stats["times"].items():
        assert v >= waste * n_steps


@mark.parametrize(["keep_order"], [(True,), (False,)])
def test_neg_sync_loop(keep_order):
    # from pprint import pprint

    n_outcomes, n_steps = 10, 10
    waste_center = 0.1
    c1 = MySyncController(sleep_seconds=waste_center, name="c1")
    c2 = MySyncController(sleep_seconds=waste_center, name="c2")
    mechanisms = []
    for m in range(2):
        mechanism = SAOMechanism(
            outcomes=n_outcomes,
            n_steps=n_steps,
            ignore_negotiator_exceptions=False,
            name=f"{m}",
        )
        ufuns = MappingUtilityFunction.generate_random(
            2, outcomes=mechanism.discrete_outcomes()
        )
        mechanism.add(
            c1.create_negotiator(preferences=ufuns[0], id=f"0-{m}", name=f"0-{m}")  # type: ignore It will be SAOControlled
        )
        mechanism.add(
            c2.create_negotiator(preferences=ufuns[1], id=f"1-{m}", name=f"1-{m}")  # type: ignore It will be SAOControlled
        )
        mechanisms.append(mechanism)
    SAOMechanism.runall(mechanisms, keep_order=keep_order)

    for mechanism in mechanisms:
        assert mechanism.state.started
        assert mechanism.state.agreement is None
        assert not mechanism.state.has_error
        assert not mechanism.state.broken
        assert mechanism.state.timedout
        assert mechanism.state.step == n_steps
        assert not mechanism.state.waiting
        assert len(mechanism.history) == n_steps
        # assert c1.n_counter_all_calls == n_steps
        # assert c2.n_counter_all_calls == n_steps


@mark.parametrize(["n_negotiators"], [(1,), (2,), (3,)])
def test_neg_run_sync(n_negotiators):
    n_outcomes, n_steps = 10, 10
    waste_edge, waste_center = 0.2, 0.1
    c = MySyncController(sleep_seconds=waste_center)
    mechanisms, edge_names = [], []
    for _ in range(n_negotiators):
        mechanism = SAOMechanism(
            outcomes=n_outcomes,
            n_steps=n_steps,
            ignore_negotiator_exceptions=True,
        )
        ufuns = MappingUtilityFunction.generate_random(
            2, outcomes=mechanism.discrete_outcomes()
        )
        edge_names.append(f"f{0}")
        mechanism.add(
            TimeWaster(
                name=f"agent{0}",
                id=f"a{0}",
                sleep_seconds=waste_edge,
                preferences=ufuns[0],
            )
        )
        mechanism.add(c.create_negotiator(preferences=ufuns[1]))  # type: ignore IT will be the correct type
        mechanisms.append(mechanism)
    SAOMechanism.runall(mechanisms)

    for mechanism in mechanisms:
        assert mechanism.state.started
        assert mechanism.state.agreement is None
        assert not mechanism.state.has_error, f"{mechanism.state.error_details}"
        assert not mechanism.state.broken
        assert mechanism.state.timedout, f"Did not timeout!!\n{mechanism.state}"
        assert mechanism.state.step == n_steps
        assert not mechanism.state.waiting
        assert len(mechanism.history) == n_steps
        delay_center = 0.0
        for k, v in mechanism.stats["times"].items():
            if k in edge_names:
                assert v >= waste_edge * n_steps
            else:
                delay_center += v
        assert delay_center > n_steps * waste_center
        assert c.n_counter_all_calls == n_steps


def test_exceptions_are_saved():
    n_outcomes, n_negotiators = 10, 2
    mechanism = SAOMechanism(
        outcomes=n_outcomes, n_steps=n_outcomes, ignore_negotiator_exceptions=True
    )
    ufuns = MappingUtilityFunction.generate_random(
        n_negotiators, outcomes=mechanism.discrete_outcomes()
    )
    mechanism.add(AspirationNegotiator(name=f"agent{0}"), preferences=ufuns[0])
    mechanism.add(MyRaisingNegotiator(name=f"agent{1}"), preferences=ufuns[1])
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism.state.step == 1
    mechanism.step()
    mechanism.step()
    assert mechanism.state.current_offer is not None
    assert len(mechanism.stats) == 3
    stats = mechanism.stats
    assert "times" in stats.keys()
    assert "exceptions" in stats.keys()
    assert stats["exceptions"] is not None
    assert len(stats["exceptions"]) == 1
    # print(stats["exceptions"][mechanism.negotiators[1].id])
    assert len(stats["exceptions"][mechanism.negotiators[1].id]) == 1
    assert exception_str in stats["exceptions"][mechanism.negotiators[1].id][0]
    assert len(stats["exceptions"][mechanism.negotiators[0].id]) == 0


def test_on_negotiation_start():
    mechanism = SAOMechanism(outcomes=10, n_steps=10)
    assert mechanism.on_negotiation_start()


@mark.parametrize(["n_negotiators"], [(2,), (3,)])
def test_round_n_agents(n_negotiators):
    n_outcomes = 5
    mechanism = SAOMechanism(outcomes=n_outcomes, n_steps=3)
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[i])
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism.state.step == 1
    assert mechanism.state.current_offer is not None


@mark.parametrize(["n_negotiators"], [(2,), (3,)])
def test_mechanism_can_run(n_negotiators):
    n_outcomes = 5
    mechanism = SAOMechanism(outcomes=n_outcomes, n_steps=3)
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[i])
    assert mechanism.state.step == 0
    mechanism.step()
    mechanism.run()


@mark.parametrize("n_negotiators,oia", [(2, True), (3, True), (2, False), (3, False)])
def test_mechanism_runs_with_offering_not_accepting(n_negotiators, oia):
    n_outcomes = 5
    mechanism = SAOMechanism(outcomes=n_outcomes, n_steps=3, offering_is_accepting=oia)
    ufuns = MappingUtilityFunction.generate_random(1, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[0])
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism._current_proposer and mechanism._current_proposer.name == "agent0"
    assert mechanism.state.n_acceptances == n_negotiators + int(oia) - 1
    assert (mechanism.agreement is not None) is oia
    if mechanism.agreement is not None:
        return
    mechanism.step()
    assert mechanism._current_proposer.name == "agent0"
    assert mechanism.state.n_acceptances == n_negotiators
    assert mechanism.agreement is not None


@mark.parametrize("n_negotiators,oia", [(2, True), (3, True), (2, False), (3, False)])
def test_mechanism_runall(n_negotiators, oia):
    n_outcomes = 5
    mechanisms = []
    for _ in range(10):
        mechanism = SAOMechanism(
            outcomes=n_outcomes,
            n_steps=random.randint(3, 20),
            offering_is_accepting=oia,
        )
        ufuns = MappingUtilityFunction.generate_random(1, outcomes=n_outcomes)
        for i in range(n_negotiators):
            mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[0])
        mechanisms.append(mechanism)

    states = SAOMechanism.runall(mechanisms)
    assert len(states) == 10
    assert not any(_ is not None and _.running for _ in states)


class MySAOSync(SAOSyncController):
    def counter_all(
        self, offers: dict[str, Outcome], states: dict[str, SAOState]
    ) -> dict[str, SAOResponse]:
        responses = {}
        for source in offers.keys():
            _, state = offers[source], states[source]
            if state.step < 2:
                responses[source] = SAOResponse(ResponseType.REJECT_OFFER, None)
            else:
                responses[source] = SAOResponse(ResponseType.ACCEPT_OFFER, None)
        return responses


@mark.parametrize(
    "n_steps,n_negotiators,oia",
    [
        (2, 2, True),
        (2, 3, True),
        (30, 2, True),
        (30, 3, True),
        (2, 2, False),
        (2, 3, False),
        (30, 2, False),
        (30, 3, False),
    ],
)
def test_random_offer_guaranteed_acceptance(n_steps, n_negotiators, oia):
    n_outcomes = 10
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=n_steps,
        time_limit=None,
        offering_is_accepting=oia,
        end_on_no_response=False,
    )
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for u in ufuns:
        u.reserved_value = float(u.min()) - 0.1
    for i in range(n_negotiators):
        mechanism.add(
            RandomAlwaysAcceptingNegotiator(name=f"agent{i}"), preferences=ufuns[i]
        )

    state = mechanism.run()
    assert state.agreement


@mark.parametrize(
    "n_negs,n_negotiators,oia",
    [
        (2, 2, True),
        (2, 3, True),
        (3, 2, True),
        (1, 2, True),
        (1, 3, True),
        (2, 2, False),
        (2, 3, False),
        (3, 2, False),
        (1, 2, False),
        (1, 3, False),
    ],
)
def test_sync_controller(n_negs, n_negotiators, oia):
    n_outcomes = 2

    mechanisms = []
    controller = MySAOSync()
    for i in range(n_negs):
        mechanisms.append(
            SAOMechanism(
                outcomes=n_outcomes,
                n_steps=30,
                time_limit=None,
                offering_is_accepting=oia,
                end_on_no_response=False,
            )
        )
        ufuns = MappingUtilityFunction.generate_random(
            n_negotiators, outcomes=n_outcomes
        )
        for u in ufuns:
            u.reserved_value = float(u.min()) - 0.1
        for i in range(n_negotiators):
            mechanisms[-1].add(
                RandomAlwaysAcceptingNegotiator(name=f"agent{i}"), preferences=ufuns[i]
            )

        mechanisms[-1].add(controller.create_negotiator())

    states = SAOMechanism.runall(mechanisms)
    assert all(_ is not None and _.agreement is not None for _ in states), str(
        [_.agreement if _ else "No State" for _ in states]
    )


def test_pickling_mechanism(tmp_path):
    import dill as pickle

    file = tmp_path / "mechanism.pck"
    n_outcomes, n_negotiators = 5, 3
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=3,
        offering_is_accepting=True,
    )
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[i])

    assert mechanism.state.step == 0
    with open(file, "wb") as f:
        pickle.dump(mechanism, f)
    with open(file, "rb") as f:
        pickle.load(f)
    assert mechanism.state.step == 0
    mechanism.step()
    with open(file, "wb") as f:
        pickle.dump(mechanism, f)
    with open(file, "rb") as f:
        pickle.load(f)
    assert mechanism.state.step == 1


@mark.xfail(
    run=False,
    reason="Checkpointing is known to fail with UtilityInverter. As this whole thing will be changed, we may just wait for now",
)
def test_checkpointing_mechanism(tmp_path):
    file = tmp_path
    n_outcomes, n_negotiators = 5, 3
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=3,
        offering_is_accepting=True,
    )
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[i])

    assert mechanism.state.step == 0
    file_name = mechanism.checkpoint(file)

    info = SAOMechanism.checkpoint_info(file_name)
    assert isinstance(info["time"], str)
    assert info["step"] == 0
    assert info["type"].endswith("SAOMechanism")
    assert info["id"] == mechanism.id
    assert info["name"] == mechanism.name

    mechanism, info = SAOMechanism.from_checkpoint(file_name, return_info=True)
    assert isinstance(info["time"], str)
    assert info["step"] == 0
    assert info["type"].endswith("SAOMechanism")
    assert info["id"] == mechanism.id
    assert info["name"] == mechanism.name
    assert isinstance(mechanism, SAOMechanism)

    assert mechanism.state.step == 0
    mechanism.step()

    file_name = mechanism.checkpoint(file)

    info = SAOMechanism.checkpoint_info(file_name)
    assert isinstance(info["time"], str)
    assert info["step"] == 1
    assert info["type"].endswith("SAOMechanism")
    assert info["id"] == mechanism.id
    assert info["name"] == mechanism.name

    mechanism, info = SAOMechanism.from_checkpoint(file_name, return_info=True)
    assert isinstance(mechanism, SAOMechanism)
    assert isinstance(info["time"], str)
    assert info["step"] == 1
    assert info["type"].endswith("SAOMechanism")
    assert info["id"] == mechanism.id
    assert info["name"] == mechanism.name

    mechanism.run()


@mark.parametrize(["n_negs"], [(_,) for _ in range(1, 10)])
def test_sync_controller_gets_all_offers(n_negs):
    from negmas.mechanisms import Mechanism
    from negmas.sao import RandomNegotiator, SAORandomSyncController

    class MyController(SAORandomSyncController):
        def counter_all(self, offers, states):
            assert len(offers) == len(self.active_negotiators)
            responses = super().counter_all(offers, states)
            assert len(responses) == len(offers)
            return responses

    c = MyController()
    negs = tuple(
        SAOMechanism(issues=[make_issue((0.0, 1.0), "price")], n_steps=20)
        for _ in range(n_negs)
    )
    for neg in negs:
        neg.add(RandomNegotiator())
        neg.add(c.create_negotiator())

    Mechanism.runall(negs, True)


# @given(n_negs=st.integers(1, 10), strict=st.booleans())
@mark.parametrize(
    ["n_negs", "strict"], [(_, s) for s in (True, False) for _ in range(1, 10)]
)
def test_single_agreement_gets_one_agreement(n_negs, strict):
    from negmas.mechanisms import Mechanism
    from negmas.sao import AspirationNegotiator, SAOSingleAgreementRandomController

    os = make_os([make_issue((0.0, 1.0), "price")])
    c = SAOSingleAgreementRandomController(strict=strict)

    negs = tuple(
        SAOMechanism(
            outcome_space=os,
            n_steps=50,
            end_on_no_response=False,
        )
        for _ in range(n_negs)
    )
    for i, neg in enumerate(negs):
        neg.add(
            AspirationNegotiator(aspiration_type="linear", name=f"opponent-{i}"),
            preferences=LinearUtilityFunction(weights=[1.0], outcome_space=os),
        )
        neg.add(c.create_negotiator(name=f"against-{i}"))

    Mechanism.runall(negs, True)
    agreements = [neg.state.agreement for neg in negs]
    if strict:
        # assert that the controller never had multiple agreements
        assert sum(_ is not None for _ in agreements) == 1
    else:
        # assert that the controller never accepted twice. It may still have multiple agreements though
        assert (
            len(
                [
                    neg.state.agreement
                    for neg in negs
                    if neg.state.agreement is not None
                    and neg.state.current_proposer is not None
                    and neg.state.current_proposer.startswith("opponent")
                ]
            )
            < 2
        )


@mark.parametrize(["keep_order"], [(False,), (True,)])
def test_loops_are_broken(keep_order):
    """Tests that loops formed by concurrent negotiations are broken for syn controllers"""
    from negmas.mechanisms import Mechanism
    from negmas.sao import SAOSingleAgreementAspirationController

    issues = [make_issue((0.0, 1.0), "price")]

    a, b, c = (
        SAOSingleAgreementAspirationController(
            preferences=MappingUtilityFunction(lambda x: x[0], issues=issues),
            strict=False,
        ),
        SAOSingleAgreementAspirationController(
            preferences=MappingUtilityFunction(lambda x: x[0], issues=issues),
            strict=False,
        ),
        SAOSingleAgreementAspirationController(
            preferences=MappingUtilityFunction(lambda x: x[0], issues=issues),
            strict=False,
        ),
    )

    n1 = SAOMechanism(
        name="ab",
        issues=issues,
        n_steps=50,
    )
    n2 = SAOMechanism(
        name="ac",
        issues=issues,
        n_steps=50,
    )
    n3 = SAOMechanism(
        name="bc",
        issues=issues,
        n_steps=50,
        end_on_no_response=False,
    )

    n1.add(a.create_negotiator(name="a>b"))
    n1.add(b.create_negotiator(name="b>a"))
    n2.add(a.create_negotiator(name="a>c"))
    n2.add(c.create_negotiator(name="c>a"))
    n3.add(b.create_negotiator(name="b>c"))
    n3.add(c.create_negotiator(name="c>b"))
    negs = (n1, n2, n3)
    Mechanism.runall(negs, keep_order)

    agreements = [neg.state.agreement for neg in negs]
    assert not keep_order or sum(_ is not None for _ in agreements) > 0
    # TODO check why sometimes we get no agreements when order is not kept


@given(
    typ=st.sampled_from(NEGTYPES),
)
def test_can_create_all_negotiator_types(typ):
    issues = [make_issue((0.0, 1.0), name="price"), make_issue(10, name="quantity")]
    params = dict(
        preferences=LinearUtilityFunction(
            issues=issues, weights=dict(price=1.0, quantity=1.0)
        )
    )
    assert typ(**params) is not None


@given(
    a=st.sampled_from(NEGTYPES),
    b=st.sampled_from(NEGTYPES),
    w1p=st.floats(-1.0, 1.0),
    w1q=st.floats(-1.0, 1.0),
    w2p=st.floats(-1.0, 1.0),
    w2q=st.floats(-1.0, 1.0),
    r1=st.floats(-1.0, 1.0),
    r2=st.floats(-1.0, 1.0),
)
@settings(deadline=100_000, max_examples=50)
def test_can_run_all_negotiators(a, b, w1p, w1q, w2p, w2q, r1, r2):
    issues = [make_issue((0.0, 1.0), name="price"), make_issue(10, name="quantity")]
    u1 = LinearUtilityFunction(weights=[w1p, w1q], issues=issues, reserved_value=r1)
    u2 = LinearUtilityFunction(weights=[w2p, w2q], issues=issues, reserved_value=r2)
    m = SAOMechanism(n_steps=30, issues=issues)
    m.add(a(preferences=u1))
    m.add(b(), preferences=u2)
    m.run()
    assert not m.running


def test_can_run_asp_tit():
    b, a = AspirationNegotiator, NaiveTitForTatNegotiator
    issues = [make_issue((0.0, 1.0), name="price"), make_issue(10, name="quantity")]
    u1 = LinearUtilityFunction(weights=[0.0, 0.0], issues=issues, reserved_value=0.0)
    u2 = LinearUtilityFunction(weights=[0.0, 0.0], issues=issues, reserved_value=0.0)
    m = SAOMechanism(n_steps=30, issues=issues)
    m.add(a(preferences=u1))
    m.add(b(), preferences=u2)
    m.run()
    assert not m.running


def test_acceptable_outcomes():
    p = SAOMechanism(outcomes=6, n_steps=10)
    p.add(
        LimitedOutcomesNegotiator(name="seller", acceptable_outcomes=[(2,), (3,), (5,)])
    )
    p.add(
        LimitedOutcomesNegotiator(name="buyer", acceptable_outcomes=[(1,), (4,), (3,)])
    )
    state = p.run()
    assert state.agreement == (3,), f"{p.extended_trace}"


class MyNegotiator(SAONegotiator):
    def propose(self, state):
        _ = state
        return (3.0, 2, 1.0)

    def respond(self, state, source=None):
        _ = source
        if state.step < 5:
            return ResponseType.REJECT_OFFER
        return ResponseType.ACCEPT_OFFER


def test_cast_offers_tuple():
    issues = [make_issue(10), make_issue(5), make_issue(3)]
    m = SAOMechanism(
        issues=issues,
        check_offers=True,
        enforce_issue_types=True,
        cast_offers=True,
        n_steps=10,
        end_on_no_response=True,
    )
    m.add(MyNegotiator())
    m.add(MyNegotiator())
    m.run()
    assert m.agreement is not None
    assert all(isinstance(_, int) for _ in m.agreement)
    assert not any(isinstance(_, float) for _ in m.agreement)
    assert m.agreement == (3, 2, 1)


def test_fail_on_incorrect_types_tuple_or_dict():
    issues = [make_issue(10), make_issue(5), make_issue(3)]
    m = SAOMechanism(
        issues=issues,
        check_offers=True,
        enforce_issue_types=True,
        cast_offers=False,
        n_steps=10,
        end_on_no_response=True,
    )
    m.add(MyNegotiator())
    m.add(MyNegotiator())
    m.run()
    assert m.agreement is None


def test_no_check_offers_tuple():
    issues = [make_issue(10), make_issue(5), make_issue(3)]
    for a, b in ((True, False), (False, False), (False, True), (True, True)):
        m = SAOMechanism(
            issues=issues,
            check_offers=False,
            enforce_issue_types=a,
            cast_offers=b,
            n_steps=10,
            end_on_no_response=True,
        )
        m.add(MyNegotiator())
        m.add(MyNegotiator())
        m.run()
        assert m.agreement is not None
        assert isinstance(m.agreement[0], float) and not isinstance(m.agreement[0], int)
        assert not isinstance(m.agreement[1], float) and isinstance(m.agreement[1], int)
        assert isinstance(m.agreement[2], float) and not isinstance(m.agreement[2], int)
        assert m.agreement == (3.0, 2, 1.0)


def test_no_limits_raise_warning():
    from negmas.inout import load_genius_domain_from_folder

    with pytest.warns(UserWarning):
        folder_name = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/cameradomain"
        )

        load_genius_domain_from_folder(
            folder_name
        ).normalize().to_single_issue().make_session()


@given(
    n_steps=st.integers(10, 11),
    n_waits=st.integers(0, 4),
    n_waits2=st.integers(0, 4),
)
@settings(deadline=20000, max_examples=100)
def test_single_mechanism_history_with_waiting(n_steps, n_waits, n_waits2):
    n_outcomes, waste = 5, (0.0, 0.3)
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=n_steps,
        ignore_negotiator_exceptions=False,
    )
    assert mechanism.outcomes
    ufuns = MappingUtilityFunction.generate_random(2, outcomes=mechanism.outcomes)
    mechanism.add(
        TimeWaster(
            name=f"agent{0}", sleep_seconds=waste, preferences=ufuns[0], n_waits=n_waits
        )
    )
    mechanism.add(
        TimeWaster(
            name=f"agent{1}",
            sleep_seconds=waste,
            preferences=ufuns[1],
            n_waits=n_waits2,
        )
    )
    mechanism.run()
    first = mechanism._selected_first
    assert first == 0
    assert mechanism.state.agreement is None
    assert mechanism.state.started
    assert mechanism.state.timedout or (
        n_waits + n_waits2 > 0 and mechanism.state.broken
    )
    assert mechanism.state.step == n_steps or (
        n_waits + n_waits2 > 0
        and mechanism.state.broken
        and mechanism.state.step <= n_steps
    )
    assert not mechanism.state.waiting
    assert len(mechanism.history) == n_steps or (
        n_waits + n_waits2 > 0
        and mechanism.state.broken
        and len(mechanism.history) <= n_steps
    )

    # check history details is correct
    s = [defaultdict(int), defaultdict(int)]
    r = [defaultdict(int), defaultdict(int)]
    h = [defaultdict(int), defaultdict(int)]
    first_offers = []
    for i, n in enumerate(mechanism.negotiators):
        first_offers.append(n.received_offers[0] is None)

        # sent and received match
        for j, w in n.my_offers.items():
            # cannot send mutlipe offers in the same step
            assert j not in s[i].keys()
            # cannot send None
            assert w is not None  # or (j == 0 and not avoid_ultimatum)
            s[i][j] = w[0]
        for j, w in n.received_offers.items():
            # cannot receive multiple ofers in the same step
            assert j not in r[i].keys()
            # canont be asked to start offering except in the first step
            assert w is not None or j == 0
            # this is the first agent to offer, ignore its first step
            if first == i and j == 0:
                # if this is the first agent, its first thing recieved must be None
                assert w is None
                continue
            assert w is not None, f"None outcome agent {i} @ {j} first is {first}"
            r[i][j] = w[0]

    assert any(first_offers) and not all(first_offers)

    # reconstruct history
    neg_map = dict(zip((_.id for _ in mechanism.negotiators), [0, 1]))
    for state in mechanism.history:
        assert isinstance(state, SAOState)
        for _, w in state.new_offers:
            a = neg_map[_]
            # cannot see the same step twice in the history of an agent
            assert state.step not in h[a].keys()
            assert w is not None
            h[a][state.step] = w[0]

    # no gaps in steps and all step sets start with 0 or 1
    for i in range(len(mechanism.negotiators)):
        for x in (r, s, h):
            steps = list(x[i].keys())
            if not steps:
                continue
            assert steps[0] in (0, 1)
            for _ in range(len(steps) - 1):
                assert steps[_] + 1 == steps[_ + 1]

    # pprint([s, r, h])
    # history matches what is stored inside agents
    for i, n in enumerate(mechanism.negotiators):
        for j, w in s[i].items():
            assert j in r[1 - i] or (j == 0)
        for j, w in r[i].items():
            assert j in s[1 - i]

    # s and r will not have matched indices but should have matched values
    s = [list(_.values()) for _ in s]
    r = [list(_.values()) for _ in r]
    h2 = [list(_ for _ in _.values()) for _ in h]
    # history matches what is stored inside agents
    for i, n in enumerate(mechanism.negotiators):
        for j, w in enumerate(s[i]):
            if j < len(r[1 - i]):
                assert r[1 - i][j] == w
            assert h2[i][j] == w

        for j, w in enumerate(r[i]):
            assert s[1 - i][j] == w


@pytest.mark.skip(reason="Known Bug")
@given(
    keep_order=st.booleans(),
    n_first=st.integers(1, 6),
    n_second=st.integers(1, 6),
    end_on_no_response=st.booleans(),
)
@settings(deadline=20000)
def test_neg_sync_loop_receives_all_offers(
    keep_order, n_first, n_second, end_on_no_response
):
    # from pprint import pprint

    n_outcomes, n_steps = 100, 10
    waste_center = 0.01
    c1s = [
        MySyncController(sleep_seconds=waste_center, name="c1") for _ in range(n_first)
    ]
    c2s = [
        MySyncController(sleep_seconds=waste_center, name="c2") for _ in range(n_second)
    ]
    mechanisms = tuple(
        SAOMechanism(
            outcomes=n_outcomes,
            n_steps=n_steps,
            ignore_negotiator_exceptions=False,
            end_on_no_response=end_on_no_response,
            name=f"{i}v{j}",
        )
        for i in range(n_first)
        for j in range(n_second)
    )
    for mechanism in mechanisms:
        assert mechanism.outcomes
        ufuns = MappingUtilityFunction.generate_random(2, outcomes=mechanism.outcomes)
        i, j = tuple(int(_) for _ in mechanism.name.split("v"))
        mechanism.add(
            c1s[i].create_negotiator(
                preferences=ufuns[0], id=f"l{i}>{j}", name=f"l{i}>{j}"
            )
        )
        mechanism.add(
            c2s[j].create_negotiator(
                preferences=ufuns[1], id=f"r{j}>{i}", name=f"r{j}>{i}"
            )
        )

    while True:
        states = SAOMechanism.stepall(mechanisms, keep_order=keep_order)
        if all(not _.running for _ in states):
            break

    for mechanism in mechanisms:
        ls = [len(mechanism.negotiator_offers(n.id)) for n in mechanism.negotiators]
        check.less_equal(abs(ls[0] - ls[1]), 1, str(mechanism.trace))
        check.equal(len(mechanism.offers), 2 * n_steps)
        for n in mechanism.negotiators:
            check.equal(len(mechanism.negotiator_offers(n.id)), n_steps)

    for c, n_partners in itertools.chain(
        zip(c1s, itertools.repeat(n_second)), zip(c2s, itertools.repeat(n_first))
    ):
        steps = set(range(1, n_steps))
        countered_steps = set(c.countered_offers.keys())
        missing_steps = steps.difference(countered_steps)
        check.equal(len(missing_steps), 0, f"Missing steps are {missing_steps}")
        for s in steps:
            partners = c.countered_offers[s]
            check.equal(
                len(partners),
                n_partners,
                f"partners {len(partners)} (of {n_partners}) step {s}.\n{partners}",
            )

    for mechanism in mechanisms:
        check.is_true(mechanism.state.started)
        check.is_none(mechanism.state.agreement)
        check.is_false(mechanism.state.has_error)
        check.is_false(mechanism.state.broken)
        check.is_false(mechanism.state.waiting)
        check.is_true(mechanism.state.timedout)
        check.equal(mechanism.state.step, n_steps)
        check.equal(len(mechanism.history), n_steps)


@given(
    n_outcomes=st.integers(5, 10),
    n_negotiators=st.integers(2, 4),
    n_steps=st.integers(1, 4),
)
@settings(deadline=None)
def test_times_are_calculated(n_outcomes, n_negotiators, n_steps):
    mechanism = SAOMechanism(outcomes=n_outcomes, n_steps=8)
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(AspirationNegotiator(name=f"agent{i}"), preferences=ufuns[i])
    assert mechanism.state.step == 0
    _strt = time.perf_counter()
    for _ in range(n_steps):
        # print(f"Stepping: {_}")
        mechanism.step()
    time.sleep(0.01)
    duration = time.perf_counter() - _strt
    # assert mechanism.current_step == n_steps
    assert mechanism.state.current_offer is not None
    assert len(mechanism.stats) == 3
    stats = mechanism.stats
    assert "round_times" in stats.keys()
    assert 0 < sum(stats["round_times"]) < duration
    assert "times" in stats.keys()
    assert "exceptions" in stats.keys()
    assert stats["times"] is not None
    assert stats["exceptions"] is not None
    assert len(stats["times"]) == n_negotiators
    assert len(stats["exceptions"]) == 0
    for i in range(n_negotiators):
        assert 0 < stats["times"][mechanism.negotiators[i].id] < duration
        assert len(stats["exceptions"][mechanism.negotiators[i].id]) == 0
    assert 0 < sum(stats["times"].values()) < duration


@given(
    n_negotiators=st.integers(2, 4),
    n_issues=st.integers(1, 3),
    presort=st.booleans(),
    stochastic=st.booleans(),
)
@settings(deadline=20000, max_examples=100)
@example(n_negotiators=2, n_issues=1, presort=False, stochastic=False)
def test_aspiration_continuous_issues(n_negotiators, n_issues, presort, stochastic):
    issues = [make_issue(values=(0.0, 1.0), name=f"i{i}") for i in range(n_issues)]
    for _ in range(5):
        mechanism = SAOMechanism(
            issues=issues,
            n_steps=10,
        )
        ufuns = [
            LinearUtilityFunction(
                issues=issues,
                weights=[3.0 * random.random(), 2.0 * random.random()],
                reserved_value=0.0,
            )
            for _ in range(n_negotiators)
        ]
        i = 0
        assert mechanism.add(
            AspirationNegotiator(
                name=f"agent{i}",
                presort=presort,
                stochastic=stochastic,
                preferences=ufuns[i],
            )
        ), "Cannot add negotiator"
        for i in range(1, n_negotiators):
            assert mechanism.add(
                AspirationNegotiator(
                    name=f"agent{i}",
                    presort=presort,
                    stochastic=stochastic,
                ),
                preferences=ufuns[i],
            ), "Cannot add negotiator"
        assert mechanism.state.step == 0
        agents = dict(zip([_.id for _ in mechanism.negotiators], mechanism.negotiators))
        mechanism.run()
        for neg, neg_id in zip(mechanism.negotiators, mechanism.negotiator_ids):
            assert neg_id in agents.keys()
            utils = [neg.ufun(_) for _ in mechanism.negotiator_offers(neg_id)]
            if presort:
                if stochastic:
                    assert all(
                        utils[_] <= utils[0] + neg.tolerance for _ in range(len(utils))
                    ), f"{utils}"
                else:
                    assert all(
                        utils[i] <= utils[j] + neg.tolerance
                        for i, j in zip(range(1, len(utils)), range(0, len(utils) - 1))
                    ), f"{utils}"


@given(
    single_checkpoint=st.booleans(),
    checkpoint_every=st.integers(0, 6),
    exist_ok=st.booleans(),
)
@settings(
    deadline=20000,
    max_examples=100,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
def test_auto_checkpoint(tmp_path, single_checkpoint, checkpoint_every, exist_ok):
    import shutil

    new_folder: Path = tmp_path / unique_name("empty", sep="")
    new_folder.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(new_folder)
    new_folder.mkdir(parents=True, exist_ok=True)
    filename = "mechanism"

    n_outcomes, n_negotiators = 5, 3
    n_steps = 50
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=n_steps,
        offering_is_accepting=True,
        checkpoint_every=checkpoint_every,
        checkpoint_folder=new_folder,
        checkpoint_filename=filename,
        extra_checkpoint_info=None,
        exist_ok=exist_ok,
        single_checkpoint=single_checkpoint,
    )
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(
            AspirationNegotiator(
                name=f"agent{i}",
                aspiration_type="conceder",
            ),
            preferences=ufuns[i],
        )

    mechanism.run()

    if 0 < checkpoint_every <= n_steps:
        if single_checkpoint:
            assert len(list(new_folder.glob("*"))) == 2
        else:
            assert len(list(new_folder.glob("*"))) >= 2 * (
                max(1, mechanism.state.step // checkpoint_every)
            )
    elif checkpoint_every > n_steps:
        assert len(list(new_folder.glob("*"))) == 2
    else:
        assert len(list(new_folder.glob("*"))) == 0


@pytest.mark.skipif(
    condition=not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_in_sao_with_time_limit_and_nsteps_raises_warning():
    from negmas.genius import GeniusNegotiator
    from negmas.inout import load_genius_domain_from_folder

    with pytest.warns(UserWarning, match=".*has a .*"):
        folder_name = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/cameradomain"
        )

        d = load_genius_domain_from_folder(folder_name)
        mechanism = d.make_session(n_steps=60, time_limit=180)
        a1 = GeniusNegotiator(
            java_class_name="agents.anac.y2017.ponpokoagent.PonPokoAgent",
            domain_file_name=d.outcome_space.name,
            utility_file_name=d.ufuns[0].name,
        )
        mechanism.add(a1)


@pytest.mark.skipif(
    condition=not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_in_sao_with_time_limit_or_nsteps_raises_no_warning():
    from negmas.genius import GeniusNegotiator
    from negmas.inout import load_genius_domain_from_folder

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        folder_name = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/cameradomain"
        )
        d = load_genius_domain_from_folder(folder_name)
        mechanism = d.make_session(n_steps=None, time_limit=180)
        a1 = GeniusNegotiator(
            java_class_name="agents.anac.y2017.ponpokoagent.PonPokoAgent",
            domain_file_name=d.outcome_space.name,
            utility_file_name=d.ufuns[0].name,
        )
        mechanism.add(a1)

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        folder_name = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/cameradomain"
        )
        d = load_genius_domain_from_folder(folder_name)
        mechanism = d.make_session(n_steps=60, time_limit=None)
        a1 = GeniusNegotiator(
            java_class_name="agents.anac.y2017.ponpokoagent.PonPokoAgent",
            domain_file_name=d.outcome_space.name,
            utility_file_name=d.ufuns[0].name,
        )
        mechanism.add(a1)


def make_mapping():
    outcome_space = make_os([make_issue(50)])
    outcomes = list(outcome_space.enumerate_or_sample())
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    return u1, u2, outcome_space


def make_linear():
    issues = [make_issue((1, 5)), make_issue((1, 10))]
    outcome_space = make_os(issues)
    u1 = LinearUtilityFunction([-0.75, 0.25], outcome_space=outcome_space)
    u2 = LinearUtilityFunction([0.5, -0.5], outcome_space=outcome_space)
    return u1, u2, outcome_space


def _run_neg(agents, utils, outcome_space):
    neg = SAOMechanism(outcome_space=outcome_space, n_steps=200, time_limit=None)
    for a, u in zip(agents, utils):
        neg.add(a, preferences=u)
    neg.run()
    assert neg.state.agreement is not None, f"No agreement!!\n{neg.trace}"
    for a, u in zip(agents, utils):
        offers = neg.negotiator_offers(a.id)
        _, best = u.extreme_outcomes()
        assert (
            isinstance(a, ConcederTBNegotiator)
            or isinstance(a, AdditiveParetoFollowingTBNegotiator)
            or isinstance(a, MultiplicativeParetoFollowingTBNegotiator)
            or u(offers[0]) >= (u(best) - 1e-4)
        ), f"Did not start with its best offer {best} (u = {u(best)}) but used {offers[0]} (u = {u(offers[0])})"
        if isinstance(a, AspirationNegotiator):
            for i, offer in enumerate(_ for _ in offers):
                assert i == 0 or u(offer) <= u(offers[i - 1]), f"Not always conceding"
                # if not(i == 0 or u(offer) <= u(offers[i - 1])):
                #     import matplotlib.pyplot as plt
                #     neg.plot()
                #     plt.show()
                #     raise AssertionError("Not always conceding")
    return neg


@given(
    typ=st.sampled_from(TIME_BASED_NEGOTIATORS),
    linear=st.booleans(),
    starting=st.booleans(),
    opp=st.sampled_from((None, NaiveTitForTatNegotiator)),
)
@settings(deadline=10_000, max_examples=10)
def test_bilateral_timebased(typ, linear, starting, opp):
    if opp is None and starting:
        return
    if opp is None:
        opp = typ
    if starting:
        a1 = typ(name=f"{typ.__name__}[0]")
        a2 = opp(name=f"{opp.__name__}[1]")
    else:
        a1 = opp(name=f"{opp.__name__}[0]")
        a2 = typ(name=f"{typ.__name__}[1]")
    u1, u2, outcome_space = make_linear() if linear else make_mapping()
    _run_neg((a1, a2), (u1, u2), outcome_space)


def test_bilateral_timebased_example():
    typ = negmas.sao.negotiators.timebased.TimeBasedNegotiator
    linear = False
    starting = False
    opp = None

    if opp is None and starting:
        return
    if opp is None:
        opp = typ
    if starting:
        a1 = typ(name=f"{typ.__name__}[0]")
        a2 = opp(name=f"{opp.__name__}[1]")
    else:
        a1 = opp(name=f"{opp.__name__}[0]")
        a2 = typ(name=f"{typ.__name__}[1]")
    u1, u2, outcome_space = make_linear() if linear else make_mapping()
    _run_neg((a1, a2), (u1, u2), outcome_space)


@given(
    neg_types=st.lists(st.sampled_from(TIME_BASED_NEGOTIATORS), min_size=4, max_size=4)
)
@settings(deadline=10_000, max_examples=20)
def test_multilateral_timebased(neg_types):
    outcome_space = make_os([make_issue(10)])
    outcomes = list(outcome_space.enumerate_or_sample())

    negotiators = [typ(name=f"{typ.__name__}") for typ in neg_types]
    utils = [
        MappingUtilityFunction(
            dict(zip(outcomes, np.random.rand(len(outcomes)).tolist())),
            outcomes=outcomes,
        )
        for _ in neg_types
    ]
    _run_neg(negotiators, utils, outcome_space)


def try_negotiator(
    cls, replace_buyer=True, replace_seller=True, plot=False, n_steps=20
):
    if isinstance(cls, str):
        cls = get_class(cls)
    from negmas.preferences import LinearAdditiveUtilityFunction
    from negmas.preferences.value_fun import AffineFun, IdentityFun, LinearFun

    buyer_cls = cls if replace_buyer else AspirationNegotiator
    seller_cls = cls if replace_seller else AspirationNegotiator

    # create negotiation agenda (issues)
    issues = [
        make_issue(name="price", values=10),
        make_issue(name="quantity", values=(1, 11)),
        make_issue(name="delivery_time", values=10),
    ]

    # create the mechanism
    session = SAOMechanism(issues=issues, n_steps=n_steps)

    # define ufuns
    seller_utility = LinearAdditiveUtilityFunction(
        values={
            "price": IdentityFun(),
            "quantity": LinearFun(0.2),
            "delivery_time": AffineFun(-1, bias=9),
        },
        weights={"price": 1.0, "quantity": 1.0, "delivery_time": 10.0},
        outcome_space=session.outcome_space,
        reserved_value=0.0,
    )
    buyer_utility = LinearAdditiveUtilityFunction(
        values={
            "price": AffineFun(-1, bias=9.0),
            "quantity": LinearFun(0.2),
            "delivery_time": IdentityFun(),
        },
        outcome_space=session.outcome_space,
        reserved_value=0.0,
    )

    session.add(
        buyer_cls(name=f"buyer-{shorten(buyer_cls.__name__)}"), ufun=buyer_utility
    )
    session.add(
        seller_cls(name=f"seller-{shorten(buyer_cls.__name__)}"), ufun=seller_utility
    )
    session.run()
    if plot:
        session.plot()
    return session


# @pytest.mark.parametrize("negotiator", [MultiplicativeParetoFollowingTBNegotiator])
@given(negotiator=st.sampled_from(ALL_BUILTIN_NEGOTIATORS))
@example(negotiator=NiceNegotiator)
@example(negotiator=NaiveTitForTatNegotiator)
@settings(deadline=100_000)
def test_specific_negotiator_buy_selling(negotiator):
    try_negotiator(negotiator, plot=False)
    # import matplotlib.pyplot as plt
    # plt.show(block=True)


if __name__ == "__main__":
    from rich import print

    print(ALL_BUILTIN_NEGOTIATORS)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

from negmas.events import Event
from negmas.helpers import exception2str, instantiate
from negmas.helpers.inout import add_records
from negmas.mechanisms import Mechanism
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue, outcome2dict

from .common import PROTOCOL_CLASS_NAME_FIELD, NegotiationInfo

if TYPE_CHECKING:
    from .agent import Agent
    from .world import World

__all__ = ["MechanismFactory"]


class MechanismFactory:
    """A mechanism creation class. It can invite agents to join a mechanism and then run it."""

    def __init__(
        self,
        world: World,
        mechanism_name: str,
        mechanism_params: dict[str, Any],
        issues: list[Issue],
        req_id: str,
        caller: Agent,
        partners: list[Agent],
        roles: list[str] | None = None,
        annotation: dict[str, Any] | None = None,
        neg_n_steps: int | None = None,
        neg_time_limit: int | None = None,
        neg_step_time_limit=None,
        allow_self_negotiation=False,
        log_ufuns_file=None,
        group: str | None = None,
    ):
        self.mechanism_name, self.mechanism_params = mechanism_name, mechanism_params
        self.caller = caller
        self.group = group
        self.partners = partners
        self.roles = roles
        self.annotation = annotation
        self.neg_n_steps = neg_n_steps
        self.neg_time_limit = neg_time_limit
        self.neg_step_time_limit = neg_step_time_limit
        self.world = world
        self.req_id = req_id
        self.issues = issues
        self.mechanism = None
        self.allow_self_negotiation = allow_self_negotiation
        self.log_ufuns_file = log_ufuns_file

    def _create_negotiation_session(
        self,
        mechanism: Mechanism,
        responses: Iterable[tuple[Negotiator, str]],
        partners: list[Agent],
    ) -> Mechanism:
        for partner in partners:
            mechanism.register_listener(event_type="negotiation_end", listener=partner)

        ufun = []
        if self.log_ufuns_file is not None:
            for outcome in mechanism.discrete_outcomes():
                record = {
                    "mechanism_id": mechanism.id,
                    "outcome": outcome2dict(outcome, mechanism.issues),
                }
                ufun.append(record)
        for i, (partner_, (_negotiator, _role)) in enumerate(zip(partners, responses)):
            if self.log_ufuns_file is not None:
                for record in ufun:
                    record[f"agent{i}"] = partner_.name
                    # record[f"agent_type{i}"] = partner_.type_name
                    # record[f"negotiator{i}"] = _negotiator.name
                    record[f"reserved{i}"] = _negotiator.reserved_value
                    try:
                        if not _negotiator.ufun:
                            record[f"u{i}"] = None
                        else:
                            record[f"u{i}"] = _negotiator.ufun(record["outcome"])
                    except:
                        record[f"u{i}"] = None
            _negotiator.owner = partner_
            mechanism.add(negotiator=_negotiator, role=_role)

        if self.log_ufuns_file is not None:
            for record in ufun:
                outcome = record.pop("outcome", {})
                record.update(outcome)
            add_records(self.log_ufuns_file, ufun)

        return mechanism

    def _start_negotiation(
        self,
        mechanism_name,
        mechanism_params: dict[str, Any],
        roles,
        caller,
        partners,
        annotation,
        issues,
        req_id,
    ) -> NegotiationInfo | None:
        """Tries to prepare the negotiation to start by asking everyone to join"""
        mechanisms = self.world.mechanisms
        if (
            (not self.allow_self_negotiation)
            and (len({_.id if _ is not None else "" for _ in partners}) < 2)
            and len(partners) > 1
        ):
            return None
        if issues is None:
            self.world.call(
                caller, caller.on_neg_request_rejected_, req_id=req_id, by=None
            )
            return None
        if (
            mechanisms is not None
            and mechanism_name is not None
            and mechanism_name not in mechanisms.keys()
        ):
            self.world.call(
                caller, caller.on_neg_request_rejected_, req_id=req_id, by=None
            )
            return None
        if mechanisms is not None and mechanism_name is not None:
            mechanism_name = mechanisms[mechanism_name].pop(
                PROTOCOL_CLASS_NAME_FIELD, mechanism_name
            )
        if mechanism_params is None:
            mechanism_params = {}
        if mechanisms and mechanisms.get(mechanism_name, None) is not None:
            mechanism_params.update(mechanisms[mechanism_name])
        # mechanism_params = {k: v for k, v in mechanism_params.items() if k != PROTOCOL_CLASS_NAME_FIELD}
        mechanism_params["n_steps"] = self.neg_n_steps
        mechanism_params["time_limit"] = self.neg_time_limit
        mechanism_params["step_time_limit"] = self.neg_step_time_limit
        mechanism_params["issues"] = issues
        mechanism_params["annotation"] = annotation
        mechanism_params["name"] = "-".join(_.id for _ in partners)
        if mechanism_name is None:
            if mechanisms is not None and len(mechanisms) == 1:
                mechanism_name = list(mechanisms.keys())[0]
            else:
                mechanism_name = "negmas.sao.SAOMechanism"
            if mechanisms and mechanisms.get(mechanism_name, None) is not None:
                mechanism_params.update(mechanisms[mechanism_name])
        try:
            mechanism = instantiate(class_name=mechanism_name, **mechanism_params)
        except Exception as e:
            s_ = exception2str()
            self.world.mechanism_exceptions[self.world.current_step].append(s_)
            self.world.agent_exceptions[caller.id].append((self.world.current_step, s_))
            mechanism = None
            self.world.logerror(
                f"Failed to create {mechanism_name} with params {mechanism_params}",
                Event("mechanism-creation-exception", dict(exception=e)),
            )
        self.mechanism = mechanism
        if mechanism is None:
            return None

        if self.mechanism:
            self.mechanism.register_listener("negotiator_exception", self.world)
        if roles is None:
            roles = [None] * len(partners)

        partner_names = [p.id for p in partners]
        responses = [
            self.world.call(
                partner,
                partner.respond_to_negotiation_request_,
                initiator=caller.id,
                partners=partner_names,
                issues=issues,
                annotation=annotation,
                role=role,
                mechanism=mechanism.nmi,
                req_id=req_id if partner == caller else None,
            )
            for role, partner in zip(roles, partners)
        ]
        if not all(responses):
            rejectors = [p for p, response in zip(partners, responses) if not response]
            rej = [_.id for _ in rejectors]
            for r in rej:
                self.world.neg_requests_rejected[r] += 1
            self.world.call(
                caller, caller.on_neg_request_rejected_, req_id=req_id, by=rej
            )
            for partner, response in zip(partners, responses):
                if partner.id != caller.id and response:
                    self.world.call(
                        partner, partner.on_neg_request_rejected_, req_id=None, by=rej
                    )
            self.world.loginfo(
                f"{caller.name} request was rejected by {[_.name for _ in rejectors]}",
                Event(
                    "negotiation-request-rejected",
                    dict(
                        req_id=req_id,
                        caller=caller,
                        partners=partners,
                        rejectors=rejectors,
                        annotation=annotation,
                    ),
                ),
            )
            return NegotiationInfo(
                mechanism=None,
                partners=partners,
                annotation=annotation,
                issues=issues,
                rejectors=rejectors,
                requested_at=self.world.current_step,
                caller=caller,
                group=self.group,
            )
        mechanism = self._create_negotiation_session(
            mechanism=mechanism, responses=zip(responses, roles), partners=partners  # type: ignore
        )
        neg_info = NegotiationInfo(
            mechanism=mechanism,
            partners=partners,
            annotation=annotation,
            issues=issues,
            requested_at=self.world.current_step,
            caller=caller,
            group=self.group,
        )
        self.world.call(
            caller,
            caller.on_neg_request_accepted_,
            req_id=req_id,
            mechanism=mechanism.nmi,
        )
        for partner, response in zip(partners, responses):
            if partner.id != caller.id:
                self.world.call(
                    partner,
                    partner.on_neg_request_accepted_,
                    req_id=None,
                    mechanism=mechanism,
                )
        self.world.loginfo(
            f"{caller.name} request was accepted",
            Event(
                "negotiation-request-accepted",
                dict(
                    req_id=req_id,
                    caller=caller,
                    partners=partners,
                    mechanism=mechanism,
                    annotation=annotation,
                ),
            ),
        )
        return neg_info

    def init(self) -> NegotiationInfo | None:
        return self._start_negotiation(
            mechanism_name=self.mechanism_name,
            mechanism_params=self.mechanism_params,
            roles=self.roles,
            caller=self.caller,
            partners=self.partners,
            annotation=self.annotation,
            issues=self.issues,
            req_id=self.req_id,
        )



import warnings
from typing import Callable

from negmas.config import CONFIG_KEY_WARN_SLOW_OPS, negmas_config

__all__ = [
    "warn",
    "warn_if_slow",
    "deprecated",
    "NegmasWarning",
    "NegmasCannotStartNegotiation",
    "NegmasIgnoredValueWarning",
    "NegmasBridgePathWarning",
    "NegmasBridgeProcessWarning",
    "NegmasInfiniteNegotiationWarning",
    "NegmasStepAndTimeLimitWarning",
    "NegmasBrdigeParsingWarning",
    "NegmasSarializationWarning",
    "NegmasIOWarning",
    "NegmasShutdownWarning",
    "NegmasWorldConfigWarning",
    "NegmasImportWarning",
    "NegmasUnusedValueWarning",
    "NegmasMemoryWarning",
    "NegmasSpeedWarning",
    "NegmasCaughtExceptionWarning",
    "NegmasVisualizationWarning",
    "NegmasNoResponseWarning",
    "NegmasLoggingWarning",
    "NegmasNumericWarning",
    "NegmasDoubleAssignmentWarning",
    "NegmasUnexpectedValueWarning",
]


class NegmasWarning(UserWarning):
    ...


def warn(message, category=NegmasWarning, stacklevel=2, source=None):
    """Issues a warning to the user. Defaults to `NegmasWarning` and stacklevel of 2."""
    return warnings.warn(message, category, stacklevel, source)


def deprecated(message):
    """Issues a deprecation warning"""
    return warnings.warn(
        message, category=DeprecationWarning, stacklevel=2, source=None
    )


class NegmasBridgePathWarning(NegmasWarning):
    ...


class NegmasBridgeProcessWarning(NegmasWarning):
    ...


class NegmasInfiniteNegotiationWarning(NegmasWarning):
    ...


class NegmasStepAndTimeLimitWarning(NegmasWarning):
    ...


class NegmasCannotStartNegotiation(NegmasWarning):
    ...


class NegmasBrdigeParsingWarning(NegmasWarning):
    ...


class NegmasSarializationWarning(NegmasWarning):
    ...


class NegmasIOWarning(NegmasWarning):
    ...


class NegmasShutdownWarning(NegmasWarning):
    ...


class NegmasWorldConfigWarning(NegmasWarning):
    ...


class NegmasImportWarning(NegmasWarning, ImportWarning):
    ...


class NegmasUnusedValueWarning(NegmasWarning):
    ...


class NegmasMemoryWarning(NegmasWarning):
    ...


class NegmasCaughtExceptionWarning(NegmasWarning):
    ...


class NegmasVisualizationWarning(NegmasWarning):
    ...


class NegmasNoResponseWarning(NegmasWarning):
    ...


class NegmasLoggingWarning(NegmasWarning):
    ...


class NegmasNumericWarning(NegmasWarning):
    ...


class NegmasSpeedWarning(NegmasWarning):
    ...


class NegmasIgnoredValueWarning(NegmasWarning):
    ...


class NegmasDoubleAssignmentWarning(NegmasWarning):
    ...


class NegmasUnexpectedValueWarning(NegmasWarning):
    ...


class NegmasSlowOperation(NegmasWarning):
    ...


def warn_if_slow(
    size: int | float,
    message: str = "Slow Operation Warning",
    op: Callable[[float], float] = lambda x: x,
) -> None:
    """
    Issues a warning for slow operations.

    Args:
        size: The size of this operation to be checked
        message: The message to print in the warning if the size is too large
        op: An operation to apply to the config value CONFIG_KEY_WARN_SLOW_OPS before comparison.

    Remarks:
        - The warning will be issued if ```op(size) >= config[CONFIG_KEY_WARN_SLOW_OPS]```
        - Default is to do no comparison (indicated by config[CONFIG_KEY_WARN_SLOW_OPS] <= 0)
        - See `negams_config` for details about getting the config value.
    """
    limit = float(negmas_config(CONFIG_KEY_WARN_SLOW_OPS, 0))
    if limit < 0.5:
        return
    nops = op(size)
    if nops < limit:
        return
    warnings.warn(
        f"{message}: n.ops={int(nops):,} (n={int(size):,}), limit={int(limit):,}",
        NegmasSlowOperation,
        2,
        None,
    )



from __future__ import annotations

import pytest

from negmas.genius import genius_bridge_is_running
from negmas.inout import Scenario
from negmas.outcomes import make_issue
from negmas.outcomes.outcome_space import make_os
from negmas.preferences import LinearAdditiveUtilityFunction as U
from negmas.sao import (
    AspirationNegotiator,
    NaiveTitForTatNegotiator,
    RandomAlwaysAcceptingNegotiator,
    RandomNegotiator,
)
from negmas.sao.mechanism import SAOMechanism
from negmas.situated.neg import Condition
from negmas.tournaments.neg import (
    cartesian_neg_tournement,
    neg_tournament,
    scenarios_from_list,
)
from tests.switches import NEGMAS_FASTRUN

SKIP_CONDITION = NEGMAS_FASTRUN


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="skipped for fast runs",
)
def test_can_run_cartesian_tournament_n_reps():
    n = 2
    issues = (
        make_issue([f"q{i}" for i in range(10)], "quantity"),
        make_issue([f"p{i}" for i in range(5)], "price"),
    )
    ufuns = [
        (
            U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
            U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
        )
        for _ in range(n)
    ]
    scenarios = [
        Scenario(
            outcome_space=make_os(issues, name=f"S{i}"),
            ufuns=u,
            mechanism_type=SAOMechanism,  # type: ignore
            mechanism_params=dict(),
        )
        for i, u in enumerate(ufuns)
    ]
    results = cartesian_neg_tournement(
        competitors=[RandomNegotiator, AspirationNegotiator],
        non_competitors=[NaiveTitForTatNegotiator],
        scenarios=scenarios,
        neg_time_limit=10,
        neg_n_steps=None,
        n_repetitions=4,
        verbose=True,
    )
    print(results)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="skipped for fast runs",
)
def test_can_run_cartesian_tournament():
    n = 2
    issues = (
        make_issue([f"q{i}" for i in range(10)], "quantity"),
        make_issue([f"p{i}" for i in range(5)], "price"),
    )
    ufuns = [
        (
            U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
            U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
        )
        for _ in range(n)
    ]
    scenarios = [
        Scenario(
            outcome_space=make_os(issues, name=f"S{i}"),
            ufuns=u,
            mechanism_type=SAOMechanism,  # type: ignore
            mechanism_params=dict(),
        )
        for i, u in enumerate(ufuns)
    ]
    results = cartesian_neg_tournement(
        competitors=[RandomNegotiator, AspirationNegotiator],
        non_competitors=[NaiveTitForTatNegotiator],
        scenarios=scenarios,
        neg_time_limit=10,
        neg_n_steps=None,
        n_steps=4,
        verbose=True,
    )
    print(results)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="skipped for fast runs",
)
def test_can_run_world_repeated_runs():
    issues = (make_issue(10, "quantity"), make_issue(5, "price"))
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [RandomNegotiator, RandomAlwaysAcceptingNegotiator]

    ufuns = (
        U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
        U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
    )

    scenarios = []

    for index in range(2):
        for partner in competitors:
            scenarios.append(
                Condition(
                    name="d0",
                    issues=issues,
                    ufuns=ufuns,
                    partner_types=(partner,),
                    index=index,
                )
            )

    print(
        neg_tournament(
            n_configs=2 * 2,
            scenarios=scenarios_from_list(scenarios),
            competitors=competitors,
            n_steps=2,
            neg_n_steps=10,
            neg_time_limit=None,
            parallelism="serial",
        )
    )


@pytest.mark.skip(reason="no way of currently testing this")
def test_can_run_tournament():
    issues = (make_issue(10, "quantity"), make_issue(5, "price"))
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [RandomNegotiator, RandomAlwaysAcceptingNegotiator]

    domains = []
    for index in range(2):
        for partner in competitors:
            domains.append(
                Condition(
                    name="d0",
                    issues=issues,
                    ufuns=(
                        U.random(issues, reserved_value=(0.0, 0.2), normalized=False),  # type: ignore
                        U.random(issues, reserved_value=(0.0, 0.2), normalized=False),  # type: ignore
                    ),
                    partner_types=(partner,),
                    index=index,
                )
            )

    neg_tournament(
        n_configs=len(domains),
        scenarios=scenarios_from_list(domains),
        competitors=competitors,
        n_steps=1,
        neg_n_steps=10,
        neg_time_limit=None,
        compact=True,
    )


@pytest.mark.skip(reason="no way of currently testing this")
def test_can_run_tournament_from_generator():
    from negmas.tournaments.neg import random_discrete_scenarios

    n_configs = 1
    n_repetitions = 1
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [RandomNegotiator, RandomAlwaysAcceptingNegotiator]

    domains = random_discrete_scenarios(issues=[5, 4, (3, 5)], partners=competitors)

    neg_tournament(
        n_configs=len(competitors) * n_configs,
        scenarios=domains,
        competitors=competitors,
        n_steps=n_repetitions,
        neg_n_steps=10,
        neg_time_limit=None,
        compact=True,
    )



from __future__ import annotations

import os
from pathlib import Path

import papermill as pm
import pytest

NEGMAS_IGNORE_TEST_NOTEBOOKS = os.environ.get("NEGMAS_IGNORE_TEST_NOTEBOOKS", False)
# NEGMAS_IGNORE_TEST_NOTEBOOKS = True


def notebooks():
    base = Path(__file__).parent.parent.parent / "notebooks"
    return list(_ for _ in base.glob("**/*.ipynb") if "checkpoints" not in str(_))


@pytest.mark.skipif(
    condition=NEGMAS_IGNORE_TEST_NOTEBOOKS,
    reason="No Genius Bridge, skipping genius-agent tests",
)
@pytest.mark.parametrize("notebook", notebooks())
def test_notebook(notebook):
    base = Path(__file__).parent.parent.parent / "notebooks"
    dst = notebook.relative_to(base)
    dst = Path(__file__).parent / "tmp_notebooks" / str(dst)
    dst.parent.mkdir(exist_ok=True, parents=True)
    pm.execute_notebook(notebook, dst)


if __name__ == "__main__":
    print(notebooks())



from __future__ import annotations

"""Implements Event management"""
import json
import random
import time
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Iterable

from negmas import warnings

from .outcomes import Issue
from .serialization import serialize
from .types import NamedObject

__all__ = [
    "Event",
    "EventSource",
    "EventSink",
    "EventLogger",
    "Notification",
    "Notifier",
    "Notifiable",
]


@dataclass
class Event:
    """An event that can be raised and consumed"""

    __slots__ = ["type", "data"]
    type: str
    data: Any


class EventSource:
    """An object capable of raising events"""

    def __init__(self):
        super().__init__()
        self.__sinks: dict[str | None, list[EventSink]] = defaultdict(list)

    def announce(self, event: Event):
        """Raises an event and informs all event sinks that are registered for notifications
        on this event type"""

        sinks = self.__sinks.get(event.type, []) + self.__sinks.get(None, [])
        random.shuffle(sinks)
        for sink in sinks:
            sink.on_event(event=event, sender=self)

    def register_listener(self, event_type: str | None, listener: EventSink):
        """Registers a listener for the given event_type.

        Args:
            event_type: The type to register. If None, the listener will
                        be registered for all types
            listener: The listening agent (must have an `on_event` method
                      that receives an event: `Event` and a sender: `EventSource`)
        """
        self.__sinks[event_type].append(listener)


class EventSink:
    """An object capable of receiving events"""

    def on_event(self, event: Event, sender: EventSource):
        pass


def myvars(x):
    if not x:
        return x
    return {
        k: v
        for k, v in vars(x).items()
        if not k.startswith("_") and not k.endswith("_")
    }


class EventLogger(EventSink):
    """
    Logs events to a file

    Args:
        file_name: Name of the file to save events into
        types: The types of events to save. If None, all events will be saved
    """

    def __init__(self, file_name: str | Path, types: list[str] | None = None):
        file_name = Path(file_name)
        file_name.parent.mkdir(parents=True, exist_ok=True)
        self._file_name = file_name
        self._types = set(types) if types else None
        self._start = time.perf_counter()

    def reset_timer(self):
        self._start = time.perf_counter()

    def on_event(self, event: Event, sender: EventSource):
        if not self._file_name:
            return
        if self._types is not None and event.type not in self._types:
            return

        def _simplify(x):
            if x is None:
                return None
            if isinstance(x, (str, int, float)):
                return x
            if isinstance(x, Issue):
                return dict(name=x.name, values=x.values)
            if isinstance(x, dict):
                return {k: _simplify(v) for k, v in x.items()}
            for y in ("id", "name"):
                if hasattr(x, y):
                    return getattr(x, y)
            if isinstance(x, Iterable):
                return list(_simplify(_) for _ in x)
            return str(x)
            # return _simplify(myvars(x))

        try:
            sid = sender.id if hasattr(sender, "id") else serialize(sender)  # type: ignore
            d = dict(
                sender=sid,
                time=time.perf_counter() - self._start,
                type=event.type,
                data=_simplify(event.data),
            )
            with open(self._file_name, "a") as f:
                f.write(f"{json.dumps(d)},\n")
        except Exception as e:
            warnings.warn(
                f"Failed to log {str(event)}: {str(e)}", warnings.NegmasLoggingWarning
            )


@dataclass
class Notification:
    __slots__ = ["type", "data"]
    type: str
    data: Any


class Notifier(NamedObject):
    """An object that can notify other objects"""

    def notify(self, notifiable: Notifiable, notification: Notification):
        notifiable.on_notification_(notification=notification, notifier=self.id)


class Notifiable:
    """An object that can be notified"""

    def add_handler(
        self, notification_type: str, callback: Callable[[Notification, str], bool]
    ):
        """
        Adds a notification handler to the list of handlers of the given type. These handlers will be called
        in the order in which they are received

        Args:
            notification_type: Notification type as specificed in the type member of the Notification class
            callback: The callback which must receive a Notification object and a string and returns a boolean. If True
                      is returned from one callback, the remaining callbacks will not be called

        Returns:

        """
        if not hasattr(self, "__notification_handlers"):
            self.__notification_handlers: dict[
                str, list[Callable[[Notification, str], bool]]
            ] = defaultdict(list)
        self.__notification_handlers[notification_type].append(callback)

    def handlers(
        self, notification_type: str
    ) -> list[Callable[[Notification, str], bool]]:
        """
        Gets the list of handlers registered for some notification type. This list can be modified in place to change
        the order of handlers for example. It is NOT a copy.
        """
        try:
            return self.__notification_handlers[notification_type]
        except (ValueError, IndexError, AttributeError):
            return []

    def remove_handler(
        self, notification_type: str, callback: Callable[[Notification, str], bool]
    ) -> bool:
        """
        Removes a notification handler from the list of handlers of the given type.

        Args:
            notification_type: Notification type as specificed in the type member of the Notification class
            callback: The callback which must receive a Notification object and a string and returns a boolean. If True
                      is returned from one callback, the remaining callbacks will not be called

        Returns:
            Whether or not the handler was in the list of handlers for this type. In all cases, the handler will not be
            called after this call (either it was not there or it will be removed).

        """
        try:
            self.__notification_handlers[notification_type].remove(callback)
            return True
        except (ValueError, IndexError, AttributeError):
            return False

    def on_notification(self, notification: Notification, notifier: str) -> None:
        """
        Called when a notification is received and is not handled by any registered handler

        Args:
            notification: The notification received
            notifier: The notifier ID

        Remarks:

            - override this method to provide a catch-all notification handling method.
        """

    def on_notification_(self, notification: Notification, notifier: str) -> bool:
        """
        Called when a notification is received. Do NOT directly override this method

        Args:
            notification:
            notifier:

        Returns:

        """
        try:
            for callback in self.__notification_handlers[notification.type]:
                if callback(notification, notifier):
                    break
            return True
        except (IndexError, ValueError, AttributeError):
            self.on_notification(notification, notifier)
            return False



from __future__ import annotations

from typing import Callable, Iterable

from negmas.generics import GenericMapping, gmap, ikeys
from negmas.helpers import get_full_type_name
from negmas.outcomes import Issue, Outcome, OutcomeRange, outcome_in_range
from negmas.outcomes.outcome_space import CartesianOutcomeSpace
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

from ..base import OutcomeUtilityMapping
from ..crisp_ufun import UtilityFunction
from ..mixins import StationaryMixin

__all__ = [
    "NonLinearAggregationUtilityFunction",
    "HyperRectangleUtilityFunction",
    "NonlinearHyperRectangleUtilityFunction",
]


class NonLinearAggregationUtilityFunction(StationaryMixin, UtilityFunction):
    r"""A nonlinear utility function.

    Allows for the modeling of a single nonlinear utility function that combines the utilities of different issues.

    Args:
        values: A set of mappings from issue values to utility functions. These are generic mappings so
                        `Callable` (s) and `Mapping` (s) are both accepted
        f: A nonlinear function mapping from a dict of utility_function-per-issue to a float
        name: name of the utility function. If None a random name will be generated.

    Notes:

        The utility is calculated as:

        .. math::

                u = f\left(u_0\left(i_0\right), u_1\left(i_1\right), ..., u_n\left(i_n\right)\right)

        where :math:`u_j()` is the utility function for issue :math:`j` and :math:`i_j` is value of issue :math:`j` in the
        evaluated outcome.


    Examples:

        >>> from negmas.outcomes import make_issue
        >>> from negmas.preferences.crisp.mapping import MappingUtilityFunction
        >>> issues = [make_issue((10.0, 20.0), 'price'), make_issue(['delivered', 'not delivered'], 'delivery')
        ...           , make_issue(5, 'quality')]
        >>> print(list(map(str, issues)))
        ['price: (10.0, 20.0)', "delivery: ['delivered', 'not delivered']", 'quality: (0, 4)']
        >>> g = NonLinearAggregationUtilityFunction({'price': lambda x: 2.0*x
        ...                                         , 'delivery': {'delivered': 10, 'not delivered': -10}
        ...                                         , 'quality': MappingUtilityFunction(lambda x: x-3)}
        ...         , f=lambda u: u[0]  + 2.0 * u[-1], issues=issues)
        >>> g((14.0, 'delivered', 2)) - ((2.0*14.0)+2.0*(2-3))
        0.0

        You must pass a value for each issue in the outcome. If some issues are not used for the ufun, you can pass them as any value that is acceptable to the corresponding value function

        >>> g = NonLinearAggregationUtilityFunction({'price'    : lambda x: 2.0*x
        ...                                         , 'delivery': {'delivered': 10, 'not delivered': -10}}
        ...         , f=lambda u: 2.0 * u[0], issues=issues[:2])
        >>> g((14.0, 'delivered')) - (2.0*(2.0*14))
        0.0

    """

    def __init__(
        self,
        values: dict[str, GenericMapping] | list[GenericMapping] | None,
        f: Callable[[tuple[float]], float],
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        if isinstance(values, dict):
            if not isinstance(self.outcome_space, CartesianOutcomeSpace):
                raise ValueError(
                    f"Cannot create a {self.__class__.__name__} with an outcome-space that is not Cartesian while passing values as a dict"
                )
            if not self.outcome_space.issues:
                raise ValueError(
                    "Cannot initializes with dict of values if you are not specifying issues"
                )
            values = [
                values[_]
                for _ in [
                    i.name if isinstance(i, Issue) else i
                    for i in self.outcome_space.issues
                ]
            ]
        self.values = values
        self.f = f

    def xml(self, issues: list[Issue]) -> str:
        raise NotImplementedError(f"Cannot convert {self.__class__.__name__} to xml")

    def to_dict(self):
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        return dict(
            **d,
            values=serialize(self.values),
            f=serialize(self.f),
        )

    @classmethod
    def from_dict(cls, d):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        for k in ("values", "f"):
            d[k] = deserialize(d.get(k, None))
        return cls(**d)

    def eval(self, offer: Outcome | None) -> float | None:
        if offer is None:
            return self.reserved_value
        if self.values is None:
            raise ValueError("No issue utilities were set.")

        u = tuple(gmap(v, w) for w, v in zip(offer, self.values))
        return self.f(u)


class HyperRectangleUtilityFunction(StationaryMixin, UtilityFunction):
    """A utility function defined as a set of hyper-volumes.

    The utility function that is calulated by combining linearly a set of *probably nonlinear* functions applied in
    predefined hyper-volumes of the outcome space.

     Args:
          outcome_ranges: The outcome_ranges for which the `mappings` are defined
          weights: The *optional* weights to use for combining the outputs of the `mappings`
          ignore_issues_not_in_input: If a hyper-volumne local function is defined for some issue
          that is not in the outcome being evaluated ignore it.
          ignore_failing_range_utilities: If a hyper-volume local function fails, just assume it
          did not exist for this outcome.
          name: name of the utility function. If None a random name will be generated.

     Examples:
         We will use the following issue space of cardinality :math:`10 \times 5 \times 4`:

         >>> from negmas.outcomes import make_issue
         >>> issues = [make_issue(10), make_issue(5), make_issue(4)]

         Now create the utility function with

         >>> f = HyperRectangleUtilityFunction(outcome_ranges=[
         ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
         ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
         ...                                , utilities= [2.0, lambda x: 2 * x[2] + x[0]])
         >>> g = HyperRectangleUtilityFunction(outcome_ranges=[
         ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
         ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
         ...                                , utilities= [2.0, lambda x: 2 * x[2] + x[0]]
         ...                                , ignore_issues_not_in_input=True)
         >>> h = HyperRectangleUtilityFunction(outcome_ranges=[
         ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
         ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
         ...                                , utilities= [2.0, lambda x: 2 * x[2] + x[0]]
         ...                                , ignore_failing_range_utilities=True)

         We can now calcualte the utility_function of some outcomes:

         * An outcome that belongs to the both outcome_ranges:
         >>> [f({0: 1.5,1: 1.5, 2: 2.5}), g({0: 1.5,1: 1.5, 2: 2.5}), h({0: 1.5,1: 1.5, 2: 2.5})]
         [8.5, 8.5, 8.5]

         * An outcome that belongs to the first hypervolume only:
         >>> [f({0: 1.5,1: 1.5, 2: 1.0}), g({0: 1.5,1: 1.5, 2: 1.0}), h({0: 1.5,1: 1.5, 2: 1.0})]
         [2.0, 2.0, 2.0]

         * An outcome that belongs to and has the first hypervolume only:
         >>> [f({0: 1.5}), g({0: 1.5}), h({0: 1.5})]
         [None, 0.0, None]

         * An outcome that belongs to the second hypervolume only:
         >>> [f({0: 1.5,2: 2.5}), g({0: 1.5,2: 2.5}), h({0: 1.5,2: 2.5})]
         [None, 6.5, None]

         * An outcome that has and belongs to the second hypervolume only:
         >>> [f({2: 2.5}), g({2: 2.5}), h({2: 2.5})]
         [None, 0.0, None]

         * An outcome that belongs to no outcome_ranges:
         >>> [f({0: 11.5,1: 11.5, 2: 12.5}), g({0: 11.5,1: 11.5, 2: 12.5}), h({0: 11.5,1: 11.5, 2: 12.5})]
         [0.0, 0.0, 0.0]


     Remarks:
         - The number of outcome_ranges, mappings, and weights must be the same
         - if no weights are given they are all assumed to equal unity
         - mappings can either by an `OutcomeUtilityMapping` or a constant.

    """

    def adjust_params(self):
        if self.weights is None:
            self.weights = [1.0] * len(self.outcome_ranges)

    def __init__(
        self,
        outcome_ranges: Iterable[OutcomeRange],
        utilities: list[float] | list[OutcomeUtilityMapping],
        weights: list[float] | None = None,
        ignore_issues_not_in_input=False,
        ignore_failing_range_utilities=False,
        bias: float = 0.0,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.outcome_ranges = list(outcome_ranges)
        self.mappings = list(utilities)
        self.weights = list(weights) if weights else ([1.0] * len(self.outcome_ranges))
        self.ignore_issues_not_in_input = ignore_issues_not_in_input
        self.ignore_failing_range_utilities = ignore_failing_range_utilities
        self.bias = bias
        self.adjust_params()

    def xml(self, issues: list[Issue]) -> str:
        """Represents the function as XML

        Args:
            issues:

        Examples:

            >>> from negmas.outcomes import make_issue
            >>> f = HyperRectangleUtilityFunction(outcome_ranges=[
            ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
            ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
            ...                                , utilities= [2.0, 9.0 + 4.0])
            >>> print(f.xml([make_issue((0.0, 4.0), name='0'), make_issue((0.0, 9.0), name='1')
            ... , make_issue((0.0, 9.0), name='2')]).strip())
            <issue index="1" name="0" vtype="real" type="real" etype="real">
                <range lowerbound="0.0" upperbound="4.0"></range>
            </issue><issue index="2" name="1" vtype="real" type="real" etype="real">
                <range lowerbound="0.0" upperbound="9.0"></range>
            </issue><issue index="3" name="2" vtype="real" type="real" etype="real">
                <range lowerbound="0.0" upperbound="9.0"></range>
            </issue><utility_function maxutility="-1.0">
                <ufun type="PlainUfun" weight="1" aggregation="sum">
                    <hyperRectangle utility_function="2.0">
                        <INCLUDES index="1" min="1.0" max="2.0" />
                        <INCLUDES index="2" min="1.0" max="2.0" />
                    </hyperRectangle>
                    <hyperRectangle utility_function="13.0">
                        <INCLUDES index="1" min="1.4" max="2.0" />
                        <INCLUDES index="3" min="2.0" max="3.0" />
                    </hyperRectangle>
                </ufun>
            </utility_function>

        """
        output = ""
        for i, issue in enumerate(issues):
            name = issue.name
            if isinstance(issue.values, tuple):
                output += (
                    f'<issue index="{i+1}" name="{name}" vtype="real" type="real" etype="real">\n'
                    f'    <range lowerbound="{issue.values[0]}" upperbound="{issue.values[1]}"></range>\n'
                    f"</issue>"
                )
            elif isinstance(issue.values, int):
                output += (
                    f'<issue index="{i+1}" name="{name}" vtype="integer" type="integer" etype="integer" '
                    f'lowerbound="0" upperbound="{issue.values - 1}"/>\n'
                )
            else:
                output += (
                    f'<issue index="{i+1}" name="{name}" vtype="integer" type="integer" etype="integer" '
                    f'lowerbound="{min(issue.values)}" upperbound="{max(issue.values)}"/>\n'
                )
        # todo find the real maxutility
        output += '<utility_function maxutility="-1.0">\n    <ufun type="PlainUfun" weight="1" aggregation="sum">\n'
        for rect, u, w in zip(self.outcome_ranges, self.mappings, self.weights):
            if not isinstance(u, float):
                raise ValueError(
                    f"Only hyper-rectangles with constant utility per rectangle can be convereted to xml"
                )
            output += f'        <hyperRectangle utility_function="{u * w}">\n'
            for key in rect.keys():
                # indx = [i for i, _ in enumerate(issues) if _.name == key][0] + 1
                indx = key + 1
                values = rect.get(key, None)
                if values is None:
                    continue
                if isinstance(values, float) or isinstance(values, int):
                    mn, mx = values, values
                elif isinstance(values, tuple):
                    mn, mx = values
                else:
                    mn, mx = min(values), max(values)
                output += (
                    f'            <INCLUDES index="{indx}" min="{mn}" max="{mx}" />\n'
                )
            output += f"        </hyperRectangle>\n"
        output += "    </ufun>\n</utility_function>"
        return output

    def to_stationary(self):
        return self

    @classmethod
    def random(
        cls,
        outcome_space,
        reserved_value,
        normalized=True,
        rectangles=(1, 4),
        **kwargs,
    ) -> HyperRectangleUtilityFunction:
        """Generates a random ufun of the given type"""
        raise NotImplementedError("random hyper-rectangle ufuns are not implemented")

    def to_dict(self):
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        d.update(super().to_dict())
        return dict(
            **d,
            outcome_ranges=serialize(self.outcome_ranges),
            utilities=serialize(self.mappings),
            weights=self.weights,
            ignore_issues_not_in_input=self.ignore_issues_not_in_input,
            ignore_failing_range_utilities=self.ignore_failing_range_utilities,
        )

    @classmethod
    def from_dict(cls, d):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        for k in ("oucome_ranges", "utilities"):
            d[k] = deserialize(d.get(k, None))
        return cls(**d)

    def eval(self, offer: Outcome | None) -> float | None:
        if offer is None:
            return self.reserved_value
        u = self.bias
        for weight, outcome_range, mapping in zip(
            self.weights, self.outcome_ranges, self.mappings
        ):  # type: ignore
            # fail on any outcome_range that constrains issues not in the presented outcome
            if (
                outcome_range is not None
                and set(ikeys(outcome_range)) - set(ikeys(offer)) != set()
            ):
                if self.ignore_issues_not_in_input:
                    continue

                return None

            elif outcome_range is None or outcome_in_range(offer, outcome_range):
                if isinstance(mapping, float):
                    u += weight * mapping
                else:
                    # fail if any outcome_range utility_function cannot be calculated from the input
                    try:
                        # noinspection PyTypeChecker
                        u += weight * gmap(mapping, offer)
                    except KeyError:
                        if self.ignore_failing_range_utilities:
                            continue

                        return None

        return u


class NonlinearHyperRectangleUtilityFunction(StationaryMixin, UtilityFunction):
    """A utility function defined as a set of outcome_ranges.


    Args:
           hypervolumes: see `HyperRectangleUtilityFunction`
           mappings: see `HyperRectangleUtilityFunction`
           f: A nonlinear function to combine the results of `mappings`
           name: name of the utility function. If None a random name will be generated
    """

    def __init__(
        self,
        hypervolumes: Iterable[OutcomeRange],
        mappings: list[OutcomeUtilityMapping],
        f: Callable[[list[float]], float],
        name: str | None = None,
        reserved_value: float = float("-inf"),
        id=None,
    ) -> None:
        super().__init__(
            name=name,
            reserved_value=reserved_value,
            id=id,
        )
        self.hypervolumes = hypervolumes
        self.mappings = mappings
        self.f = f

    def xml(self, issues: list[Issue]) -> str:
        raise NotImplementedError(f"Cannot convert {self.__class__.__name__} to xml")

    def to_dict(self):
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        d.update(super().to_dict())
        return dict(
            **d,
            hypervolumes=serialize(self.hypervolumes),
            mappings=serialize(self.mappings),
            f=serialize(self.f),
        )

    @classmethod
    def from_dict(cls, d):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        for k in ("hypervolumes", "mapoints", "f"):
            d[k] = deserialize(d.get(k, None))
        return cls(**d)

    def eval(self, offer: Outcome | None) -> float | None:
        if offer is None:
            return self.reserved_value
        if not isinstance(self.hypervolumes, Iterable):
            raise ValueError(
                "Hypervolumes are not set. Call set_params() or pass them through the constructor."
            )

        u = []
        for hypervolume, mapping in zip(self.hypervolumes, self.mappings):
            if outcome_in_range(offer, hypervolume):
                u.append(gmap(mapping, offer))
        return self.f(u)



from __future__ import annotations

import pytest

from negmas.outcomes import generate_issues
from negmas.outcomes.outcome_space import CartesianOutcomeSpace

from .fixtures import bissue, cissue, dissue, hamlet, uissue


def test_type(hamlet, cissue, dissue):
    assert cissue.type == "continuous", "Continuous type passes"
    assert dissue.type == "contiguous", "int passes"
    assert hamlet.type == "categorical", "string list passes"


def test_is_continuous(cissue, dissue, hamlet):
    assert cissue.is_continuous() is True, "Continuous type passes"
    assert dissue.is_continuous() is False, "string list passes"
    assert hamlet.is_continuous() is False, "int passes"


def test_is_discrete(cissue, dissue, hamlet):
    assert cissue.is_discrete() is False, "Continuous type passes"
    assert dissue.is_discrete() is True, "string list passes"
    assert hamlet.is_discrete() is True, "int passes"


def test_string_conversion(uissue, hamlet):
    assert str(uissue).endswith(": ['val 0', 'val 1', 'val 2', 'val 3', 'val 4']")
    assert str(hamlet) == "THE problem: ['val 0', 'val 1', 'val 2', 'val 3', 'val 4']"


def test_cartinatlity(cissue, dissue, hamlet):
    assert cissue.cardinality == float("inf")
    assert hamlet.cardinality == 5
    assert dissue.cardinality == 10


def test_n_outcomes(cissue, dissue, hamlet):
    assert CartesianOutcomeSpace([cissue, dissue, hamlet]).cardinality == float("inf")
    assert CartesianOutcomeSpace([dissue, hamlet]).cardinality == 50
    assert CartesianOutcomeSpace([dissue]).cardinality == 10
    assert CartesianOutcomeSpace([]).cardinality == 1


def test_rand(cissue, dissue, hamlet, bissue):
    for _ in range(100):
        assert 0.0 <= cissue.rand() < 1.0
        assert 0 <= dissue.rand() < 10
        assert hamlet.rand() in hamlet.values
        assert bissue.rand() in bissue.values


def test_rand_invalid(cissue, dissue, hamlet, bissue):
    for _ in range(100):
        assert not 0.0 <= cissue.rand_invalid() < 1.0
        assert not 0 <= dissue.rand_invalid() < 10
        assert hamlet.rand_invalid() not in hamlet.values
        assert bissue.rand_invalid() not in bissue.values


def test_possibilities(cissue, dissue, hamlet, bissue):
    with pytest.raises(ValueError):
        _ = list(cissue.all)
    assert len(list(dissue.all)) == 10
    assert len(list(hamlet.all)) == 5
    assert len(list(bissue.all)) == 2


def test_issue_generation_defaults():
    options = ["a", "b", "c"]
    issues = generate_issues([(0.0, 1.0), options, 5])
    assert len(issues) == 3
    assert (
        issues[0].is_continuous()
        and issues[0]._values[0] == 0.0
        and issues[0]._values[1] == 1.0
    )
    for i, o in enumerate(options):
        assert issues[1]._values[i] == o
    assert issues[2]._values == (0, 4)
    for i, issue in enumerate(issues):
        assert str(i) == issue.name


def test_issue_generation_multiples():
    issues_ = generate_issues([5], [10])
    assert len(issues_) == 10
    for i, issue in enumerate(issues_):
        assert issue.name.startswith(str(i))
        assert issue._values == (0, 4)


if __name__ == "__main__":
    pytest.main(args=[__file__])



from __future__ import annotations

from ..components.acceptance import AcceptAround, RandomAcceptancePolicy
from ..components.offering import RandomOfferingPolicy
from .modular import MAPNegotiator

__all__ = [
    "RandomNegotiator",
    "RandomAlwaysAcceptingNegotiator",
]


class RandomNegotiator(MAPNegotiator):
    """
    A negotiation agent that responds randomly in a single negotiation.

    Args:
        p_acceptance: Probability of accepting an offer
        p_rejection:  Probability of rejecting an offer
        p_ending: Probability of ending the negotiation at any round
        can_propose: Whether the agent can propose or not
        **kwargs: Passed to the GBNegotiator

    Remarks:
        - If p_acceptance + p_rejection + p_ending < 1, the rest is the probability of no-response.
    """

    def __init__(
        self,
        p_acceptance=0.15,
        p_rejection=0.75,
        p_ending=0.1,
        can_propose=True,
        **kwargs,
    ) -> None:
        acceptance = RandomAcceptancePolicy(p_acceptance, p_rejection, p_ending)
        offering = None if not can_propose else RandomOfferingPolicy()
        super().__init__(acceptance=acceptance, offering=offering, **kwargs)


class RandomAlwaysAcceptingNegotiator(MAPNegotiator):
    def __init__(
        self,
        p_acceptance=0.15,
        p_rejection=0.75,
        p_ending=0.1,
        can_propose=True,
        accept_around=1.0,
        eps=1e-3,
        **kwargs,
    ) -> None:
        acceptance = RandomAcceptancePolicy(
            p_acceptance, p_rejection, p_ending
        ) | AcceptAround(accept_around, eps=eps)
        offering = None if not can_propose else RandomOfferingPolicy()
        super().__init__(acceptance=acceptance, offering=offering, **kwargs)



"""Implements Checkpoint functionality for easy dumping and restoration of any `NamedObject` in negmas."""

from __future__ import annotations

import shutil
from os import PathLike
from pathlib import Path
from typing import Any, Callable

import numpy as np

from negmas.helpers.inout import load
from negmas.types import NamedObject


class CheckpointMixin:
    """Adds the ability to save checkpoints to a `NamedObject`"""

    def checkpoint_init(
        self,
        step_attrib: str = "current_step",
        every: int = 1,
        folder: PathLike | None = None,
        filename: str | None = None,
        info: dict[str, Any] | None = None,
        exist_ok: bool = True,
        single: bool = True,
    ):
        """
        Initializes the object to automatically save a checkpoint

        Args:
            step_attrib: The attribute that defines the current step. If None, there is no step concept
            every: Number of steps per checkpoint. If < 1 no checkpoints will be saved
            folder: The directory to store checkpoints under
            filename: Name of the file to save the checkpoint under. If None, a unique name will be chosen.
                                 If `single_checkpoint` was False, then multiple files will be used prefixed with the
                                 step number
            info: Any extra information to save in the json file associated with each checkpoint
            exist_ok: Override existing files if any
            single: If True, only the most recent checkpoint will be kept

        Remarks:

            - single_checkpoint implies exist_ok

        """
        self.__checkpoint_every = -1 if folder is None else every
        self.__checkpoint_folder = folder
        self.__checkpoint_extra_info = info
        self.__checkpoint_exist_ok = exist_ok
        self.__checkpoint_single = single
        self.__step_atrrib = step_attrib
        self.__checkpoint_filename = filename

    def checkpoint_on_step_started(self) -> Path | None:
        """Should be called on every step to save checkpoints as needed.

        Returns:
            The path on which the checkpoint is stored if one is stored. None otherwise.

        Remarks:

            - Should be called at the BEGINNING of every step before any processing takes place
        """
        if self.__checkpoint_every < 1 or self.__checkpoint_folder is None:
            return None
        step = getattr(self, self.__step_atrrib)
        if step % self.__checkpoint_every == 0 or self.__checkpoint_every == 1:
            me: NamedObject = self  # type: ignore
            return me.checkpoint(
                path=self.__checkpoint_folder,
                file_name=self.__checkpoint_filename,
                info=self.__checkpoint_extra_info,
                exist_ok=self.__checkpoint_exist_ok or self.__checkpoint_single,
                single_checkpoint=self.__checkpoint_single,
                step_attribs=(self.__step_atrrib,),
            )

    def checkpoint_final_step(self) -> Path | None:
        """Should be called at the end of the simulation to save the final state

        Remarks:
            - Should be called after all processing of the final step is conducted.
        """
        if self.__checkpoint_every < 1 or self.__checkpoint_folder is None:
            return None
        me: NamedObject = self  # type: ignore
        return me.checkpoint(
            path=self.__checkpoint_folder,
            file_name=self.__checkpoint_filename,
            info=self.__checkpoint_extra_info,
            exist_ok=True,
            single_checkpoint=self.__checkpoint_single,
            step_attribs=(self.__step_atrrib,),
        )


class CheckpointRunner:
    """Runs an object based on its checkpoints saved in an earlier run"""

    def register_callback(self, callback: Callable[[NamedObject, int], None]) -> None:
        """Registers a callback to be called whenever a new step is loaded

        Args:
            callback: A callable that takes the named object (after it is loaded and an integer specifying the step
                      number and returns None.

        """
        self.__callbacks.append(callback)

    def __init__(
        self,
        folder: str | Path,
        id: str | None = None,
        callback: Callable[[NamedObject, int], None] | None = None,
        watch: bool = False,
        object_type: type[NamedObject] = NamedObject,
    ):
        self.__folder = Path(folder).absolute()
        if id is None:
            pattern = "*.json"
        else:
            pattern = "*id*.json"
        self.__infos = [load(_) for _ in self.__folder.glob(pattern)]
        self.__files = dict(
            zip(
                (_["step"] for _ in self.__infos), (_["filename"] for _ in self.__infos)
            )
        )
        self.__sorted_steps = sorted(list(self.__files.keys()))
        self._step_index = -1
        self.__object: NamedObject | None = None
        self.__object_type = object_type
        self.__callbacks = []
        if callback is not None:
            self.register_callback(callback)
        self.__watch = watch
        if watch:
            raise NotImplementedError("File watching is not implemented yet")

    @property
    def current_step(self) -> int:
        """Gets the current step number"""
        if self._step_index < 0:
            return -1
        return self.__sorted_steps[self._step_index]

    def goto(self, step: int, exact=False) -> int | None:
        """Goes to the nearest step for the given one returning the exact step number.

        Args:

            step: The step we want to goto
            exact: If True, must go to the exact step number, otherwise go to the nearest step stored in a checkpoint

        Returns:

            - None if the current step is the nearest to the given step. Otherwise the exact step we moved to

        """
        if step is None:
            return None
        step_index = np.searchsorted(self.__sorted_steps, step, side="left")
        if step_index > 0 and self.__sorted_steps[step_index - 1] == step:
            step_index -= 1
        if not exact:
            n = len(self.__sorted_steps)
            if step_index > n:
                step = n - 1
            else:
                step = self.__sorted_steps[step_index]

        if self._step_index > -1 and step == self.__sorted_steps[self._step_index]:
            return None
        filename = self.__files.get(step)
        if not filename:
            raise ValueError(f"step {step} has no file")
        self.__object = self.__object_type.from_checkpoint(filename, return_info=False)
        self._step_index = step_index
        for callback in self.__callbacks:
            callback(self.__object, step)
        return step

    def step(self) -> int | None:
        """Go one step forward in the stored steps.

        Returns:
            The number of the current step or None if we are already on the last step.

        """
        nxt_step = self._step_index + 1
        if len(self.__sorted_steps) > nxt_step:
            return self.goto(self.__sorted_steps[nxt_step], exact=True)
        return None

    def run(self):
        """Run all steps. Notice that if `register_callback` was used to register some callback functions, they will
        be called for every stored stepped during the run."""
        while self.step() is not None:
            pass

    @property
    def loaded_object(self) -> NamedObject | None:
        """The object stored in the current checkpoint"""
        return self.__object

    def fork(
        self,
        copy_past_checkpoints: bool = False,
        every: int = 1,
        folder: str | Path | None = None,
        filename: str | None = None,
        info: dict[str, Any] | None = None,
        exist_ok: bool = True,
        single: bool = True,
    ) -> NamedObject | None:
        """
        Creates a copy of the internal object that can be run safely.

        Args:
            copy_past_checkpoints: If true, all checkpoints upto and including current_step will be copied to the given
                                   folder
            every: Number of steps per checkpoint. If < 1 no checkpoints will be saved
            folder: The directory to store checkpoints under
            filename: Name of the file to save the checkpoint under. If None, a unique name will be chosen.
                                 If `single_checkpoint` was False, then multiple files will be used prefixed with the
                                 step number
            info: Any extra information to save in the json file associated with each checkpoint
            exist_ok: Override existing files if any
            single: If True, only the most recent checkpoint will be kept

        Returns:

        """
        if self.__object is None:
            return None
        if (
            not isinstance(self.__object, CheckpointMixin)
            and folder is not None
            and every > 0
        ):
            raise ValueError(
                f"Object of type {self.__object.__class__.__name__} is not implementing the "
                f"CheckpointMixin. It cannot be forked"
            )
        if copy_past_checkpoints and folder is None:
            raise ValueError(
                "Cannot copy past checkpoints because no folder for new checkpoints is given"
            )

        if folder is None:
            folder = Path()
        else:
            folder = Path(folder).absolute()

        if copy_past_checkpoints:
            files = [v for k, v in self.__files.items() if k <= self.current_step]
            for f in files:
                shutil.copy(str(f), str(folder / Path(f).name))
                shutil.copy(str(f) + ".json", str(folder / (Path(f).name + ".json")))
        x = self.__object
        if isinstance(self.__object, CheckpointMixin):
            CheckpointMixin.checkpoint_init(
                x,  # type: ignore
                every=every,
                folder=folder,
                filename=filename,
                info=info,
                exist_ok=exist_ok,
                single=single,
            )
        return x

    @property
    def steps(self) -> list[int]:
        """A list of all stored steps"""
        return self.__sorted_steps

    def reset(self) -> None:
        """Goes before the first step"""
        self._step_index = -1
        self.__object = None

    @property
    def next_step(self) -> int | None:
        """Get the  next stored step number (None if it does not exist)"""
        nxt = self._step_index + 1
        if len(self.__sorted_steps) > nxt:
            return self.__sorted_steps[nxt]
        return None

    @property
    def previous_step(self) -> int | None:
        """Get the  previous stored step number (None if it does not exist)"""
        if self._step_index < 0:
            return -1
        nxt = self._step_index - 1
        if 0 <= nxt:
            return self.__sorted_steps[nxt]
        return None

    @property
    def last_step(self) -> int:
        """Get the  last stored step number (None if it does not exist)"""
        return self.__sorted_steps[-1]

    @property
    def first_step(self) -> int:
        """Get the  first stored step number (None if it does not exist)"""
        return self.__sorted_steps[0]



from __future__ import annotations

import shutil
from os import walk
from pathlib import Path

import pytest
from pytest import mark

from negmas.genius.bridge import genius_bridge_is_running
from negmas.inout import Scenario
from negmas.preferences.crisp.nonlinear import HyperRectangleUtilityFunction
from negmas.preferences.discounted import DiscountedUtilityFunction

MAX_CARDINALITY = 10_000

try:
    # try to raise the maximum limit for open files
    import resource

    resource.setrlimit(resource.RLIMIT_NOFILE, (50000, -1))
except:
    pass

# todo: get these to work
SCENARIOS_TO_IGNORE = [
    "S-1NIKFRT-3",
    "IntegerDomain",
    "S-1NIKFRT-2",
    "S-1NAGUNL-128",
    "S-1NAGUNL-255",
    "50issueswithRV",
    "50issues",
    "30issuesDiscounted",
    "30issueswithRV",
    "30issues",
    "50issuesDiscounted",
    "30issuesDiscountedwithRV",
    "50issuesDiscountedwithRV",
    "AgentHp2",
    "web_service",
    "four_issues",
    "AMPOvsCity",
    "laptopdomain",
    "S-1NIKFRT-1",
    "S-1NAGUNL-114",
    "laptopdomainNoBayes",
    "inheritancedomain",
    "10issues",
    "10issuesDiscounted",
    "10issueswithRV",
    "10issuesDiscountedwithRV",
    "group9-vacation",
    "group9-killer_robot",
    "FitnessA",
    "FitnessB",
    "FitnessC",
    "Grandma",
    "YXAgent",
    # "kite",
    # "group5-car_domain",
    # "group2-new_sporthal",
    # "group12-symposium",
    # "group11-car_purchase",
    # "group9-vacation",
    # "group8-holiday",
]


def get_all_scenarios():
    base = Path(__file__).parent.parent / "data" / "scenarios"
    data = []
    for root, dirs, files in walk(base):
        if len(files) == 0 or len(dirs) != 0:
            continue
        if root.split("/")[-1] in SCENARIOS_TO_IGNORE:
            continue
        data.append(root)
    return data


def compared_two_domains(domain, domain2):
    from negmas.genius.gnegotiators import AgentX, Atlas3

    assert len(domain.issues) == len(domain2.issues)
    for i1, i2 in zip(domain.issues, domain2.issues):
        assert (
            i1.cardinality == i2.cardinality
            and i1.type == i2.type
            and i1.value_type == i2.value_type
        )

    assert len(domain.ufuns) == len(domain2.ufuns)

    for u1, u2 in zip(domain.ufuns, domain2.ufuns):
        while isinstance(u1, DiscountedUtilityFunction):
            u1 = u1.ufun
        while isinstance(u2, DiscountedUtilityFunction):
            u2 = u2.ufun
        if isinstance(u1, HyperRectangleUtilityFunction) or isinstance(
            u2, HyperRectangleUtilityFunction
        ):
            continue
        dm = domain.outcome_space.to_discrete(5)
        for i, w in enumerate(dm):
            if i > MAX_CARDINALITY:
                return
            u1_, u2_ = u1(w), u2(w)
            assert isinstance(u1_, float)
            assert isinstance(u2_, float)
            assert abs(u1_ - u2_) < 1e-3, f"{str(u1)}\n{str(u2)}"

    for ufun in domain.ufuns:
        if isinstance(ufun, HyperRectangleUtilityFunction):
            continue
        m = domain.make_session(n_steps=100, name=domain.outcome_space.name)
        assert m is not None
        if not genius_bridge_is_running():
            continue
        n1 = Atlas3(domain_file_name=m.name, preferences=ufun)
        n2 = AgentX(domain_file_name=m.name, utility_file_name=ufun.name)
        m.add(n1)
        m.add(n2)
        u1, u2 = n1.ufun, n2.ufun
        assert u1 and u2
        outcomes = m.discrete_outcomes(max_cardinality=100)
        for outcome in outcomes:
            u1_, u2_ = u1(outcome), u2(outcome)
            assert isinstance(u1_, float)
            assert isinstance(u2_, float)
            assert abs(u1_ - u2_) < 1e-3
        n1.destroy_java_counterpart()
        n2.destroy_java_counterpart()


def do_enc_dec_trial(tmp, folder_name, with_discounting=True):
    domain = Scenario.from_genius_folder(folder_name, safe_parsing=False)
    assert domain
    domain = domain.remove_discounting()
    print(f"{str(folder_name)}\n-> {str(tmp)}")
    domain.to_genius_folder(tmp)
    domain2 = Scenario.from_genius_folder(tmp)
    assert domain2
    if not with_discounting:
        domain2 = domain2.remove_discounting()
    compared_two_domains(domain, domain2)
    try:
        shutil.rmtree(tmp)
    except:
        pass


@mark.xfail(
    run=False, reason="Known to fail. It is the int/discrete issue ambiguity in Genius"
)
@pytest.mark.parametrize("disc", [True, False])
def test_encoding_decoding_example_AMPOvsCity(tmp_path, disc):
    folder_name = (
        Path(__file__).parent.parent / "data" / "scenarios" / "other" / "AMPOvsCity"
    )
    do_enc_dec_trial(tmp_path / "tmp", folder_name, disc)


@pytest.mark.parametrize("disc", [True, False])
def test_encoding_decoding_example_group_8_holiday(tmp_path, disc):
    folder_name = (
        Path(__file__).parent.parent
        / "data"
        / "scenarios"
        / "anac"
        / "y2015"
        / "group8-holiday"
    )
    do_enc_dec_trial(tmp_path / "tmp", folder_name, disc)


@pytest.mark.parametrize("folder_name", get_all_scenarios())
def test_encoding_decoding_all_without_discounting(tmp_path, folder_name):
    do_enc_dec_trial(tmp_path / "tmp", folder_name, False)


@pytest.mark.parametrize("folder_name", get_all_scenarios())
def test_encoding_decoding_all_with_discounting(tmp_path, folder_name):
    do_enc_dec_trial(tmp_path / "tmp", folder_name, True)



"""
Base class for all rational agents in NegMAS.

A rational agent is one that has some preferences.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..common import PreferencesChange
from .named import NamedObject

if TYPE_CHECKING:
    from ..outcomes import Outcome
    from ..preferences import (
        BaseUtilityFunction,
        Preferences,
        ProbUtilityFunction,
        UtilityFunction,
    )

__all__ = ["Rational"]


class Rational(NamedObject):
    """
    A rational object is an object that can have preferences.


    Args:
        name: Object name. Used for printing and logging but not internally by the system
        preferences: An optional preferences to attach to the object
        ufun: An optinoal utility function (overrides preferences if given)
        id: Object ID (must be unique in the whole system). If None, the system will generate it


    Remarks:

        - Rational objects can be created with a predefined preferences or without them
          but the preferences sould be set for any kind of rational object involved in
          negotiations **before** the negotiation starts and should not be changed after
          that.
        - `ufun` is aliased to `preferences`

    """

    def _set_pref_owner(self):
        if not self._preferences:
            return
        # todo: re-enable this after making sure scml does not raise it
        # we here assume that two entities that share an ID can share preferences without warning
        # if (
        #     self._preferences.owner is not None
        #     and self._preferences.owner.id != self.id
        # ):
        #     warnings.warn(
        #         f"Entity {self.name} ({self.__class__.__name__}) is "
        #         f"assigned preferences belonging to another entity "
        #         f"({self._preferences.owner.name} of type {self.__class__.__name__})!!",
        #         warnings.NegmasDoubleAssignmentWarning,
        #     )

        self._preferences.owner = self

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        """
        Called to inform the entity that its ufun has changed.

        Args:
            changes: An ordered list of changes that happened.

        Remarks:

            - You MUST call the super() version of this function either before or after your code when you are overriding
              it.
            - The most general form of change is `PreferencesChange.General` which indicates that you cannot trust anything you knew about the ufun anymore
        """
        _ = changes

    def set_preferences(
        self, value: Preferences | None, force=False
    ) -> Preferences | None:
        """
        Sets tha utility function/Preferences.

        Args:

            value: The value to set to
            force: If true, `on_preferecnes_changed()` will always be called even if `value` == `self.preferences`

        """
        if value == self._preferences:
            if force:
                self.on_preferences_changed([PreferencesChange()])
            return self._preferences
        old = self._preferences
        self._preferences = value
        if value and value.owner != self:
            self._set_pref_owner()
        if id(value) != id(old):
            self.on_preferences_changed([PreferencesChange()])
        return self._preferences

    def __init__(
        self,
        name: str | None = None,
        preferences: Preferences | None = None,
        ufun: BaseUtilityFunction | None = None,
        id: str | None = None,
        type_name: str | None = None,
    ):
        super().__init__(name, type_name=type_name, id=id)
        if ufun:
            preferences = ufun
        self._init_preferences = preferences
        self._preferences = None
        if preferences is not None:
            self.set_preferences(preferences)

    @property
    def preferences(self) -> Preferences | None:
        """The utility function attached to that object"""
        return self._preferences

    @property
    def crisp_ufun(self) -> UtilityFunction | None:
        """Returns the preferences if it is a CrispUtilityFunction else None"""
        from negmas.preferences import UtilityFunction

        return (
            self._preferences
            if isinstance(self._preferences, UtilityFunction)
            else None
        )

    @crisp_ufun.setter
    def crisp_ufun(self, v: UtilityFunction):
        from negmas.preferences import UtilityFunction

        if not isinstance(v, UtilityFunction):
            raise ValueError(f"Cannot assign a {type(v)} to crisp_ufun")
        self.set_preferences(v)

    @property
    def prob_ufun(self) -> ProbUtilityFunction | None:
        """Returns the preferences if it is a ProbUtilityFunction else None"""
        from negmas.preferences import ProbUtilityFunction

        return (
            self._preferences
            if isinstance(self._preferences, ProbUtilityFunction)
            else None
        )

    @prob_ufun.setter
    def prob_ufun(self, v: ProbUtilityFunction):
        from negmas.preferences import ProbUtilityFunction

        if not isinstance(v, ProbUtilityFunction):
            raise ValueError(f"Cannot assign a {type(v)} to prob_ufun")
        self.set_preferences(v)

    @property
    def ufun(self) -> BaseUtilityFunction | None:
        """Returns the preferences if it is a `BaseUtilityFunction` else None"""
        from ..preferences import BaseUtilityFunction

        if self._preferences is None:
            return None
        if isinstance(self._preferences, BaseUtilityFunction):
            return self._preferences
        raise ValueError(
            f"Preferences are not for type `BaseUtilityFunction` ({self._preferences.__class__.__name__})"
        )

    @ufun.setter
    def ufun(self, v: BaseUtilityFunction):
        self.set_preferences(v)

    @property
    def has_preferences(self) -> bool:
        """Does the entity has an associated ufun?"""
        return self._preferences is not None

    @property
    def has_ufun(self) -> bool:
        """Does the entity has an associated ufun?"""
        return self.ufun is not None

    @property
    def has_cardinal_preferences(self) -> bool:
        """Does the entity has an associated ufun?"""
        from negmas.preferences.protocols import CardinalCrisp

        return self._preferences is not None and isinstance(
            self._preferences, CardinalCrisp
        )

    @property
    def reserved_outcome(self) -> Outcome | None:
        """
        Reserved outcome is the outcome that will be realized by default for this agent.

        Remarks:

            - Reserved outcomes are defined for `OrdinalPreferences`.

        See Also:
            `reserved_value`

        """
        from negmas.preferences import HasReservedOutcome

        if self._preferences is None or not isinstance(
            self._preferences, HasReservedOutcome
        ):
            return None
        return self._preferences.reserved_outcome

    @property
    def reserved_value(self) -> float:
        """
        Reserved value is what the entity gets if no agreement is reached in the negotiation.

        The reserved value can either be explicity defined for the ufun or it can be the output of the ufun
        for `None` outcome.

        """
        from negmas.preferences import HasReservedValue

        if self._preferences is None or not isinstance(
            self._preferences, HasReservedValue
        ):
            return float("nan")
        return self._preferences.reserved_value



"""
Implements serialization to and from strings and secondary storage.

"""
from __future__ import annotations

from pathlib import Path
from typing import Any, Iterable, Type

import cloudpickle
import numpy as np
from pandas import json_normalize

from negmas import warnings

from .helpers import (
    TYPE_START,
    get_class,
    get_full_type_name,
    is_jsonable,
    is_lambda_or_partial_function,
    is_not_lambda_nor_partial_function,
)
from .helpers.inout import dump, load

__all__ = [
    "serialize",
    "deserialize",
    "dump",
    "load",
    "to_flat_dict",
    "PYTHON_CLASS_IDENTIFIER",
]

PYTHON_CLASS_IDENTIFIER = "__python_class__"
PATH_START = "__PATH__:"
LAMBDA_START = b"__LAMBDAOBJ__:"
FUNCTION_START = b"__FUNCTION_START__:"
# JSON_START = b"__JSON_START__:"
CLOUDPICKLE_START = b"__CLOUDPICKLE_START__:"
SPECIAL_FIELDS = ("_NamedObject__uuid", "_NamedObject__name")
SPECIAL_FIELDS_SHORT_NAMES = ("id", "name")


def serialize(
    value,
    deep=True,
    add_type_field=True,
    keep_private=False,
    ignore_methods=True,
    ignore_lambda=False,
    shorten_type_field=False,
    objmem=None,
):
    """
    Encodes the given value as nothing more complex than simple dict
    of either dicts, lists or builtin numeric or string values. The resulting
    dictionary will be json serializable

    Args:

        value: Any object
        deep: Whether we should go deep in the encoding or do a shallow encoding
        add_type_field: Whether to add a type field. If True, A field named `PYTHON_CLASS_IDENTIFIER` will be added
        giving the type of `value`
        keep_private: Keeps fields starting with "_"
        shorten_type_field: IF given, the type will be shortened to class name only if it starts with "negmas."

    Remarks:

        - All iterables are converted to lists when `deep` is true.
        - If the `value` object has a `to_dict` member, it will be called to
          do the conversion, otherwise its `__dict__` or `__slots__` member
          will be used.

    See Also:
          `deserialize`, `PYTHON_CLASS_IDENTIFIER`

    """

    def add_to_mem(x, objmem):
        if not objmem:
            objmem = {id(x)}
        else:
            objmem.add(id(x))
        return objmem

    def good_field(k: str, v, objmem):
        if not isinstance(k, str):
            return True
        if objmem and id(v) in objmem:
            return False
        if ignore_lambda and is_lambda_or_partial_function(v):
            return False
        if ignore_methods and is_not_lambda_nor_partial_function(v):
            return False
        if not isinstance(k, str):
            return False
        return keep_private or not (k != PYTHON_CLASS_IDENTIFIER and k.startswith("_"))

    def adjust_dict(d):
        if not isinstance(d, dict):
            return d
        for a, b in zip(SPECIAL_FIELDS, SPECIAL_FIELDS_SHORT_NAMES):
            if a in d.keys():
                if b in d.keys() and d[b] != d[a]:
                    warnings.warn(
                        f"Field {a} and {b} already exist and are not equal.",
                        warnings.NegmasSarializationWarning,
                    )
                d[b] = d[a]
                del d[b]
        return d

    if value is None:
        return None

    def get_type_field(value):
        t = value.__class__.__name__
        if shorten_type_field and t.startswith("negmas."):
            return t
        return value.__class__.__module__ + "." + t

    if isinstance(value, dict):
        if not deep:
            return adjust_dict({k: v for k, v in value.items()})
        return adjust_dict(
            {
                k: serialize(v, deep=deep, add_type_field=add_type_field, objmem=objmem)
                for k, v in value.items()
                if good_field(k, v, objmem)
            }
        )
    if isinstance(value, Iterable) and not deep:
        # add_to_mem(value)
        return value
    if isinstance(value, Type):
        return TYPE_START + get_full_type_name(value)
    if isinstance(value, Path):
        return PATH_START + str(value)
    # if isinstance(value, np.ndarray):
    #     return value.tolist()
    if isinstance(value, (list, tuple)) and not isinstance(value, str):
        objmem = add_to_mem(value, objmem)
        return adjust_dict(
            type(value)(
                serialize(_, deep=deep, add_type_field=add_type_field, objmem=objmem)
                for _ in value
            )
        )
    if hasattr(value, "to_dict"):
        converted = value.to_dict()  # type: ignore
        if isinstance(converted, dict):
            if add_type_field and (PYTHON_CLASS_IDENTIFIER not in converted.keys()):
                converted[PYTHON_CLASS_IDENTIFIER] = get_type_field(value)
            return adjust_dict({k: v for k, v in converted.items()})
        else:
            return adjust_dict(converted)
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        if (
            value.startswith(FUNCTION_START)
            or value.startswith(LAMBDA_START)
            or value.startswith(CLOUDPICKLE_START)
            # or value.startswith(JSON_START)
        ):
            warnings.warn(
                f"{value} starts with a reserved part!! Will just keep it as"
                f" it is. May be you are serializing an already serialized object",
                warnings.NegmasSarializationWarning,
            )
        return value

    if is_lambda_or_partial_function(value):
        return LAMBDA_START + cloudpickle.dumps(value)

    if is_not_lambda_nor_partial_function(value):
        return FUNCTION_START + cloudpickle.dumps(value)

    if hasattr(value, "__dict__"):
        if deep:
            objmem = add_to_mem(value, objmem)
            d = {
                k: serialize(v, deep=deep, add_type_field=add_type_field, objmem=objmem)
                for k, v in value.__dict__.items()
                if good_field(k, v, objmem)
            }
        else:
            d = {k: v for k, v in value.__dict__.items() if good_field(k, v, objmem)}
        if add_type_field:
            d[PYTHON_CLASS_IDENTIFIER] = get_type_field(value)
        return adjust_dict(d)

    if hasattr(value, "__slots__"):
        if deep:
            objmem = add_to_mem(value, objmem)
            d = dict(
                zip(
                    (k for k in value.__slots__),  # type: ignore
                    (
                        serialize(
                            getattr(value, _),
                            deep=deep,
                            add_type_field=add_type_field,
                            objmem=objmem,
                        )
                        for _ in value.__slots__  # type: ignore
                    ),
                )
            )
        else:
            d = dict(
                zip(
                    (k for k in value.__slots__),  # type: ignore
                    (getattr(value, _) for _ in value.__slots__),  # type: ignore
                )
            )
        if add_type_field:
            d[PYTHON_CLASS_IDENTIFIER] = get_type_field(value)
        return adjust_dict(d)
    if isinstance(value, np.int64):  # type: ignore
        return int(value)
    # a builtin
    if is_jsonable(value):
        return value
    try:
        vv = CLOUDPICKLE_START + cloudpickle.dumps(value)
        return vv
    except:
        pass
    warnings.warn(
        f"{value} of type {type(value)} is not serializable",
        warnings.NegmasSarializationWarning,
    )
    return value


def to_flat_dict(
    value, deep=True, add_type_field=False, shorten_type_field=False
) -> dict[str, Any]:
    """
    Encodes the given value as a flat dictionary

    Args:
        value: The value to be converted to a flat dictionary
        deep: Converting all sub-objects
        add_type_field: If true, a special field for the object type will be added
        shorten_type_field: If true, the type field will be shortened to just class name if it is defined in NegMAS

    Returns:

    """
    d = serialize(
        value,
        add_type_field=add_type_field,
        shorten_type_field=shorten_type_field,
        deep=deep,
    )
    if d is None:
        return {}
    if not isinstance(d, dict):
        raise ValueError(
            f"value is of type {type(value)} cannot be converted to a flat dict"
        )
    for k, v in d.items():
        if isinstance(v, list) or isinstance(v, tuple):
            d[k] = str(v)
    return json_normalize(d, errors="ignore", sep="_").to_dict(orient="records")[0]


def deserialize(
    d: Any,
    deep=True,
    remove_type_field=True,
    keep_private=False,
    fallback_class_name: str | None = None,
):
    """Decodes a dict/object coming from `serialize`

    Args:

        d: The value to be decoded. If it is not a dict, it is returned as it is.
        deep: If true, decode recursively
        remove_type_field: If true the field called `PYTHON_CLASS_IDENTIFIER` will be removed if found.
        keep_private: If given, private fields (starting with _) will be kept
        fallback_class_name: If given, it is used as the fall-back  type if ``PYTHON_CLASS_IDENTIFIER` is not in the dict.

    Remarks:

        - If the object is not a dict or if it has no `PYTHON_CLASS_IDENTIFIER` field and no `fallback_class_name` is
          given, the input `d` is returned as it is. It will not even be copied.

    See Also:
        `serialize`, `PYTHON_CLASS_IDENTIFIER`



    """

    def good_field(k: str):
        if not isinstance(k, str):
            return True
        return keep_private or not (k != PYTHON_CLASS_IDENTIFIER and k.startswith("_"))

    if d is None or isinstance(d, int) or isinstance(d, float) or isinstance(d, str):
        return d
    if isinstance(d, dict):
        if remove_type_field:
            python_class_name = d.pop(PYTHON_CLASS_IDENTIFIER, fallback_class_name)
        else:
            python_class_name = d.get(PYTHON_CLASS_IDENTIFIER, fallback_class_name)
        if python_class_name is not None and python_class_name != "functools.partial":
            python_class = get_class(python_class_name)
            # we resolve sub-objects first from the dict if deep is specified before calling deserialize on the class
            if deep:
                d = {
                    k: deserialize(v, deep=deep) for k, v in d.items() if good_field(k)
                }
            # deserialize needs to do a shallow conversion from a dict as deep conversion is taken care of already.
            if hasattr(python_class, "from_dict"):
                return python_class.from_dict({k: v for k, v in d.items()})  # type: ignore
            if deep:
                d = {k: deserialize(v) for k, v in d.items() if good_field(k)}
            else:
                d = {k: v for k, v in d.items() if good_field(k)}
            return python_class(**d)
        if not deep:
            return d
        return {k: deserialize(v, deep=deep) for k, v in d.items() if good_field(k)}
    if not deep:
        return d
    if isinstance(d, str):
        if d.startswith(TYPE_START):
            return get_class(d[len(TYPE_START) :])
        elif d.startswith(PATH_START):
            return Path(d[len(PATH_START) :])
        return d
    if isinstance(d, bytes):
        if d.startswith(LAMBDA_START):
            return cloudpickle.loads(d[len(LAMBDA_START) :])
        if d.startswith(FUNCTION_START):
            return cloudpickle.loads(d[len(FUNCTION_START) :])
        if d.startswith(CLOUDPICKLE_START):
            return cloudpickle.loads(d[len(CLOUDPICKLE_START) :])
        # if d.startswith(JSON_START):
        #     return json.loads(d[JSON_START:])
        return d
    if isinstance(d, tuple) or isinstance(d, list):
        return type(d)(deserialize(_) for _ in d)
    return d



from __future__ import annotations

from typing import Any, Generator

from negmas.outcomes.base_issue import Issue

__all__ = ["CallableIssue"]


class CallableIssue(Issue):
    """
    An `Issue` with a callable for generating values. This is a very limited issue type and most operations are not supported on it.
    """

    def __init__(self, values, name=None) -> None:
        super().__init__(values, name)
        self._value_type = object

    def _to_xml_str(
        self,
        indx,
    ):
        raise NotImplementedError("CallableIssue cannot be saved to xml")

    @property
    def type(self) -> str:
        return "uncountable"

    def is_uncountable(self) -> bool:
        return True

    def is_continuous(self) -> bool:
        return False

    def ordered_value_generator(
        self, n: int | float | None = 10, grid=True, compact=False, endpoints=True
    ) -> Generator[Any, None, None]:
        raise NotImplementedError(
            "Cannot generate values in order from a Callable issue"
        )

    def value_generator(
        self, n: int | float | None = 10, grid=True, compact=False, endpoints=True
    ) -> Generator[Any, None, None]:
        if n is None or isinstance(n, float):
            raise ValueError("Real valued issue with no discretization value")
        yield from (self._values() for _ in range(n))

    def rand(self):
        """Picks a random valid value."""
        return self._values()

    def rand_outcomes(
        self, n: int, with_replacement=False, fail_if_not_enough=False
    ) -> list:
        if not with_replacement:
            raise ValueError(
                "'values' is specified as a callable for this issue. Cannot "
                "sample from it without replacement"
            )

        return [self._values() for _ in range(n)]

    def rand_invalid(self):
        """Pick a random *invalid* value"""

        raise ValueError(
            f"Cannot generate invalid outcomes because values is given as a callable"
        )

    def is_valid(self):
        raise ValueError("Cannot check the validity of callable issues")

    def value_at(self, index: int):
        raise ValueError("Cannot index a callable issue")



from __future__ import annotations

from collections import defaultdict
from pathlib import Path

from .common import EDGE_COLORS, EDGE_TYPES

__all__ = ["safe_min", "deflistdict", "show_edge_colors"]


def safe_min(a, b):
    """Returns min(a, b) assuming None is less than anything."""
    if a is None:
        return b
    if b is None:
        return a
    return min(a, b)


def deflistdict():
    return defaultdict(list)


def show_edge_colors():
    """Plots the edge colors used with their meaning"""
    import matplotlib.pyplot as plt

    colors = {}
    for t in EDGE_TYPES:
        colors[t] = EDGE_COLORS[t]

    # Sort colors by hue, saturation, value and name.
    # sorted_colors = colors.values()
    sorted_names = colors.keys()

    n = len(sorted_names)
    ncols = 2
    nrows = n // ncols + 1

    fig, ax = plt.subplots(figsize=(8, 5))

    # Get height and width
    X, Y = fig.get_dpi() * fig.get_size_inches()
    h = Y / (nrows + 1)
    w = X / ncols

    for i, name in enumerate(sorted_names):
        col = i % ncols
        row = i // ncols
        y = Y - (row * h) - h

        xi_line = w * (col + 0.05)
        xf_line = w * (col + 0.25)
        xi_text = w * (col + 0.3)

        ax.text(
            xi_text,
            y,
            name,
            fontsize=(h * 0.3),
            horizontalalignment="left",
            verticalalignment="center",
        )

        ax.hlines(
            y + h * 0.1, xi_line, xf_line, color=colors[name], linewidth=(h * 0.6)
        )

    ax.set_xlim(0, X)
    ax.set_ylim(0, Y)
    ax.set_axis_off()

    fig.subplots_adjust(left=0, right=1, top=1, bottom=0, hspace=0, wspace=0)
    plt.show()



#!/usr/bin/env python
"""
A set of logging utility functions that can be used by agents developed for the platform.
"""
from __future__ import annotations

import datetime
import logging
import os
import sys
from pathlib import Path

import colorlog

from negmas.config import negmas_config

__all__ = [
    "create_loggers",
]
LOGS_BASE_DIR = Path(negmas_config("log_base", Path.home() / "negmas" / "logs"))
COMMON_LOG_FILE_NAME = str(
    Path(LOGS_BASE_DIR)
    / "{}_{}.txt".format("log", datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
)

MODULE_LOG_FILE_NAME: dict[str, str] = dict()


def create_loggers(
    file_name: str | None = None,
    module_name: str | None = None,
    screen_level: int | None = logging.WARNING,
    file_level: int | None = logging.DEBUG,
    format_str: str = "%(asctime)s - %(levelname)s - %(message)s",
    colored: bool = True,
    app_wide_log_file: bool = True,
    module_wide_log_file: bool = False,
) -> logging.Logger:
    """
    Create a set of loggers to report feedback.

    The logger created can log to both a file and the screen at the  same time
    with adjustable level for each of them. The default is to log everything to
    the file and to log WARNING at least to the screen

    Args:
        module_wide_log_file:
        app_wide_log_file:
        file_name: The file to export_to the logs to. If None only the screen
                    is used for logging. If empty, a time-stamp is used
        module_name: The module name to use. If not given the file name
                    without .py is used
        screen_level: level of the screen logger
        file_level: level of the file logger
        format_str: the format of logged items
        colored: whether or not to try using colored logs

    Returns:
        logging.Logger: The logger

    """
    if module_name is None:
        module_name = __file__.split("/")[-1][:-3]
    # create logger if it does not already exist
    logger = None
    if module_wide_log_file or app_wide_log_file:
        logger = logging.getLogger(module_name)
        if len(logger.handlers) > 0:
            return logger
        logger.setLevel(logging.DEBUG)
    else:
        logger = logging.getLogger()
    # create formatter
    file_formatter = logging.Formatter(format_str)
    if colored and "colorlog" in sys.modules and os.isatty(2) and screen_level:
        date_format = "%Y-%m-%d %H:%M:%S"
        cformat = "%(log_color)s" + format_str
        screen_formatter = colorlog.ColoredFormatter(
            cformat,
            date_format,
            log_colors={
                "DEBUG": "magenta",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_red",
            },
        )
    else:
        screen_formatter = logging.Formatter(format_str)
    if screen_level is not None and (module_wide_log_file or app_wide_log_file):
        # create console handler and set level to logdebug
        screen_logger = logging.StreamHandler()
        screen_logger.setLevel(screen_level)
        # add formatter to ch
        screen_logger.setFormatter(screen_formatter)
        # add ch to logger
        logger.addHandler(screen_logger)
    if file_name is not None and file_level is not None:
        file_name = str(file_name)
        logger = logging.getLogger(file_name)
        logger.setLevel(file_level)
        if len(file_name) == 0:
            if app_wide_log_file:
                file_name = COMMON_LOG_FILE_NAME
            elif module_wide_log_file and module_name in MODULE_LOG_FILE_NAME.keys():
                file_name = MODULE_LOG_FILE_NAME[module_name]
            else:
                file_name = "{}/{}_{}.txt".format(
                    LOGS_BASE_DIR,
                    module_name,
                    datetime.datetime.now().strftime("%Y%m%d-%H%M%S"),
                )
                MODULE_LOG_FILE_NAME[module_name] = file_name

            os.makedirs(f"{LOGS_BASE_DIR}", exist_ok=True)
        os.makedirs(os.path.dirname(file_name), exist_ok=True)  # type: ignore
        file_logger = logging.FileHandler(file_name)
        file_logger.setLevel(file_level)
        file_logger.setFormatter(file_formatter)
        logger.addHandler(file_logger)
    return logger



from __future__ import annotations

import sys

from ..common import DEFAULT_JAVA_PORT
from ..helpers.misc import get_free_tcp_port

__all__ = [
    "DEFAULT_JAVA_PORT",
    "DEFAULT_PYTHON_PORT",
    "DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT",
    "ANY_JAVA_PORT",
    "RANDOM_JAVA_PORT",
    "get_free_tcp_port",
]

DEFAULT_PYTHON_PORT = 25338
RANDOM_JAVA_PORT = 0
ANY_JAVA_PORT = -1
DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT = sys.maxsize
